#ifndef TRACK_PARAMETERS_DISTRIBUTION_PROCESSOR_H
#define TRACK_PARAMETERS_DISTRIBUTION_PROCESSOR_H 1

//LCIO
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

namespace AIDA{
  class IHistogram1D;
}


namespace marlintpc
{

    class TrackFitterFactory;
//    class TrackFitterBase;

    /** Fill AIDA histograms with all track parameters:
     *  phi, lambda, tan(lambda), d0, z0.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A Collection of reconstructed tracks 
     *
     *  <h4>Output</h4>
     *  \li Histograms of the phi and lambda distributions
     *
     *  @param InputTrackes Name of input Tracks collection (default: TPCTracks)
     *  @param HistogramMinValuePhi  Minimum value in phi histogram (default: -50. mrad)
     *  @param HistogramMaxValuePhi  Maximal value in phi histogram (default: 50. mrad)
     *  @param HistogramMinValueLambda  Minimum value in lambda histogram (default: -50. mrad)
     *  @param HistogramMaxValueLambda  Maximal value in lamnda histogram (default: 50. mrad)
     *  @param HistogramMinValueTanLambda  Minimum value in tan(lambda) histogram (default: -20. )
     *  @param HistogramMaxValueTanLambda  Maximal value in tan(lamnda) histogram (default: 20. )
     *  @param HistogramMinValueD0  Minimum value in d0 histogram (default: -50. mm)
     *  @param HistogramMaxValueD0  Maximal value in d0 histogram (default: 50. mm)
     *  @param HistogramMinValueZ0  Minimum value in z0 histogram (default: -10. mm)
     *  @param HistogramMaxValueZ0  Maximal value in z0 histogram (default: 300. mm)
     *  @param HistogramMinValueOmega  Minimum value in Omega histogram (default: -0.01 mm^-1)
     *  @param HistogramMaxValueOmega  Maximal value in Omega histogram (default: 0.01. mm^-1)
     *
     *  @author Martin Killenberg, University of Bonn
     */

    class TrackParametersDistributionProcessor : public marlin::Processor{
			      
    public:
    
	virtual Processor*  newProcessor() { return new TrackParametersDistributionProcessor ; }
    
	TrackParametersDistributionProcessor();

	~TrackParametersDistributionProcessor();

	virtual void init();

	virtual void processRunHeader( EVENT::LCRunHeader* run ) ;

	virtual void processEvent( LCEvent * evt ) ; 
    
    protected:

     std::string _inputColName ;    ///< the name of the input collection
     
     AIDA::IHistogram1D * _distributionPhiHisto; ///< The histogram of the phi values
     AIDA::IHistogram1D * _distributionLambdaHisto; ///< The histogram of the lambda values
     AIDA::IHistogram1D * _distributionTanLambdaHisto; ///< The histogram of the tan(lambda) values
     AIDA::IHistogram1D * _distributionD0Histo; ///< The histogram of the d0 values
     AIDA::IHistogram1D * _distributionZ0Histo; ///< The histogram of the z0 values
     AIDA::IHistogram1D * _distributionOmegaHisto; ///< The histogram of the z0 values

     float  _histogramMinValuePhi; ///< Minimum value in phi histogram
     float  _histogramMaxValuePhi; ///< Maximum value in phi histogram
     float  _histogramMinValueLambda; ///< Minimum value in lambda histogram
     float  _histogramMaxValueLambda; ///< Maximum value in lambda histogram
     float  _histogramMinValueTanLambda; ///< Minimum value in lambda histogram
     float  _histogramMaxValueTanLambda; ///< Maximum value in lambda histogram
     float  _histogramMinValueD0; ///< Minimum value in lambda histogram
     float  _histogramMaxValueD0; ///< Maximum value in lambda histogram
     float  _histogramMinValueZ0; ///< Minimum value in lambda histogram
     float  _histogramMaxValueZ0; ///< Maximum value in lambda histogram
     float  _histogramMinValueOmega; ///< Minimum value in Omega histogram
     float  _histogramMaxValueOmega; ///< Maximum value in Omega histogram

    };
}

#endif // PHI_LAMBDA_DISTRIBUTION_PROCESSOR_H
