#include "TimePixClusterSizeProcessor.h"
#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"

//C++
#include <iostream>
#include <cmath>

using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerData.h>
#include "UTIL/LCRelationNavigator.h"
//#include "IntersectionCalculator.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
// endif

#include <marlin/Global.h>

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata;

namespace marlintpc
{

TimePixClusterSizeProcessor aTimePixClusterSizeProcessor;
    
    TimePixClusterSizeProcessor::TimePixClusterSizeProcessor(): Processor("TimePixClusterSizeProcessor") ,
								_distributionPixelHisto(0) ,
								_distributionRadiusHisto(0)//, 
								//_distributionZHisto(0), 
								//_distributionXYHisto(0)
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputData" , 
				"Name of the input data collection"  ,
				_inputColName ,
				std::string("TimePixTrackCuts") ) ;
    // modify processor description
    _description = "TimePixClusterSizeProcessor:Fill the number of pixels in a cluster "
	"and the cluster radius in a histogram " ;

  registerProcessorParameter( "HistogramMinValuePixel" , 
			      "Minimum value in x histogram (default: 0. )"  ,
			      _histogramMinValuePixel ,
			      float(0.) ) ;

  registerProcessorParameter( "HistogramMaxValuePixel" , 
			      "Maximal value in x histogram (default:  1000. )"  ,
			      _histogramMaxValuePixel ,
			      float(1000.) ) ;

  registerProcessorParameter( "HistogramMinValueRadius" , 
			      "Minimum value in y histogram (default: 0. mm)"  ,
			      _histogramMinValueRadius ,
			      float(0.) ) ;

  registerProcessorParameter( "HistogramMaxValueRadius" , 
			      "Maximal value in y histogram (default:  15. mm)"  ,
			      _histogramMaxValueRadius ,
			      float(15.) ) ;

  registerProcessorParameter( "pitch" , 
			      "pitch of the chip in mm"  ,
			      _pitch ,
			      (float) 0.055 ) ;

  registerProcessorParameter( "ZDistributionBinning" , 
			      "Binning of z distribution of histograms"  ,
			      _zDistributionBinning ,
			      int (100) ) ;

  registerProcessorParameter( "MinValueZ" , 
			      "Minimum value in z binning"  ,
			      _minValueZ ,
			      float(0.) ) ;

  registerProcessorParameter( "MaxValueZ" , 
			      "Maximal value in z binning"  ,
			      _maxValueZ ,
			      float(260.) ) ;

  registerProcessorParameter( "UnseparatedCluster" , 
			      "True if only unseparated clusters should be checked"  ,
			      _unseparatedCluster ,
			      bool(false) ) ;

}

TimePixClusterSizeProcessor::~TimePixClusterSizeProcessor()
{
}

void TimePixClusterSizeProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;
	
	int pixelBins = static_cast<int>(_histogramMaxValuePixel - _histogramMinValuePixel);
	int radiusBins = static_cast<int>( (_histogramMaxValueRadius - _histogramMinValueRadius) * 100 );
	
	for ( int i = 0; i < _zDistributionBinning; i++ ) {
	    
	    std::stringstream number;
	    number << i;
	    std::string titlePixel =  "TimePixClusterSizeHisto_" + number.str();
	    std::string titleRadius =  "TimePixClusterRadiusHisto_" + number.str();
	    
	    _distributionPixelHistoVec.push_back( AIDAProcessor::histogramFactory(this)->
						  createHistogram1D( titlePixel.c_str(), 
								     titlePixel.c_str(), 
								     pixelBins, 
								     _histogramMinValuePixel, 
								     _histogramMaxValuePixel) );

	    _distributionRadiusHistoVec.push_back( AIDAProcessor::histogramFactory(this)->
						  createHistogram1D( titleRadius.c_str(), 
								     titleRadius.c_str(), 
								     radiusBins, 
								     _histogramMinValueRadius, 
								     _histogramMaxValueRadius) );
	    
	}

	// get a histograms from the AIDAProcessor
	_distributionPixelHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "TimePixClusterSizeHisto", 
			     "TimePixClusterSizeHisto", 
			     pixelBins, _histogramMinValuePixel, _histogramMaxValuePixel);

	
	_distributionRadiusHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "TimePixClusterRadiusHisto", 
			     "TimePixClusterRadiusHisto", 
			     radiusBins, _histogramMinValueRadius, _histogramMaxValueRadius);



    }

void TimePixClusterSizeProcessor::processRunHeader( EVENT::LCRunHeader* run )
{
	
}
    
void TimePixClusterSizeProcessor::processEvent( LCEvent * evt ) 
{ 
    //read input data
    streamlog_out(MESSAGE3) << "Event: " << evt->getEventNumber() << std::endl;

    //calculate stepwidth of z binning:
    float step = ( _maxValueZ - _minValueZ ) / _zDistributionBinning;
    streamlog_out(DEBUG0) << "step = "<<step<<std::endl;

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* InputCollection=0;
    try{
	InputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	streamlog_out(DEBUG0) << "no input data returning"<<std::endl;
	return;
    }

    streamlog_out(DEBUG0) << "input data"<<std::endl;

    // cast did not succeed
    if( InputCollection == 0 )
    {
	streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
	return;
    }

      LCCollection* RelationsCol = 0;
    try {
	RelationsCol = evt->getCollection( "TimePixRelationHit" ) ;
    }
    catch(DataNotAvailableException){

//	return;
    }

    LCRelationNavigator* relationNavigator = 0; 
    if ( RelationsCol != 0 ) {
	relationNavigator = new LCRelationNavigator( RelationsCol );
	 }

    for(LCCollectionVec::iterator it = InputCollection->begin() ; 
	it < InputCollection->end(); it ++ ) {
	 streamlog_out(DEBUG0) << "test"<<std::endl;
	Track* track = dynamic_cast<Track*>(*it);
	if ( track != 0 ) {
	    
	  
	    TrackerHitVec trackerHits = track->getTrackerHits();
	    streamlog_out(DEBUG0)<<"trackerHits.size() = "<<trackerHits.size()<<std::endl;

	    for ( TrackerHitVec::iterator hit = trackerHits.begin(); hit != trackerHits.end(); hit++ ) {
		int clusterSize = 0;
		std::vector<LCObject*> dataVec;
		
		
		if ( _unseparatedCluster ) {
		    //test if this cluster arose from the separation of a bigger cluster
		    EVENT::LCObjectVec RelatedObjects = 
			relationNavigator->getRelatedFromObjects( *hit );
	    
		    if ( RelatedObjects.size() == 0 ) //cluster was not created from separation
			dataVec = (*hit)->getRawHits();
		    
		}
		else //we are not interested in whether the cluster arose from separation or not
		    dataVec = (*hit)->getRawHits();

		for ( std::vector<LCObject*>::iterator d = dataVec.begin(); d != dataVec.end(); d++ ) {
		
		    TrackerData* data = dynamic_cast<TrackerData*>( *d ) ;
		    
		    clusterSize += data->getChargeValues().size();

		}
		
		streamlog_out(DEBUG0)<<"dataVec.size() = "<<dataVec.size()<<std::endl;

		const double *pos = (*hit)->getPosition();
		


		
		_distributionPixelHisto->fill( clusterSize );


		//pi r^2 = clustersize * pitch^2
		float clusterRadius = sqrt( clusterSize * _pitch * _pitch / M_PI  );
		_distributionRadiusHisto->fill( clusterRadius );


		//fill z-binned histograms
		for ( int bin = 0; bin < _zDistributionBinning; bin++ ) {
		    
		    if ( pos[2] >= _minValueZ + bin*step && 
			 pos[2] <  _minValueZ + (bin+1)*step ) {
			
			_distributionPixelHistoVec[bin]->fill( clusterSize );
			_distributionRadiusHistoVec[bin]->fill( clusterRadius );
			
		    }
		}


	    }
	    
	}
	
    }

//	
//	_distributionZHisto->fill(pos[2]);
////	    if (pos[2] < 10)
//	_distributionTOTYHisto->fill(pos[0],pos[1]);
	
    
    
}



}// namespace 
