#ifndef PedestalCalculator_h
#define PedestalCalculator_h 1
/* PedestalCalculator.h
 * $Id: PedestalCalculator.h,v 1.1 2006/07/07 08:58:52 mjanssen Exp $
 */
// #define USE_CONDDB
//c++ stl 
#include <utility>

//LCIO
#include "lcio.h"

//MARLIN
#include "marlin/Processor.h"

//TPCCondData
#include "Pedestal.h"

//LCCD
#include <lccd.h>
#include <lccd/DBInterface.hh>

namespace marlintpc
{

    /// Calculates the pedestals out of raw data.
    /** This processor calculates the channel pedestals out of measurement raw data. 
     *  The signals are filtered out and the pedestal values and widths are 
     *  calculated from the remaining ADC spectrum.
     *  Usually, one thousand events are enough to make a decent calculation.
     *  The processor writes the pedestal in every case at the end of the input and always when a new run header comes.
     *  In addition the parameter WriteEveryNEvents can be set so that after a certain amount of events a new collection will
     *  be written.
     *  After the calculation the \link tpcconddata::Pedestal Pedestal objects \endlink are stored in a LCCollectionVector and stored as LCCD data.
     * @param InputColName Name ofName for the input colletion the input colletion (string, default="TPCConvertedRawData")
     * @param PedestalCollectionName Name of the collection containing the pedestals. (string, default: "TPCPedestal")
     * @param SimpleFileName Name of the simple LCIO file the pedestals should be written to (without: .slcio !). If this is set, the processor will write the *pedestal calculation results to a simple LCIO file in a collection. The name given here will be expanded by the run numbers (as given in the run header) and if WriteEveryNEvents is set in addition by a 3 digit counting variable. Also the file name extension ".slcio" will be automatically added (optional, string)
     * @param DBFolder Folder in the database where the pedestals will be written to. This is an optional parameter. if it is set, the skript will try to write the results to a database. If not and the simple file option is not chossen, the data will not be written after the calculation. (optional string, default "test/pedestals")
     * @param DBInitString Init spring for the connection to the database. This parameter is optional. If you do not set it, the environment variable $LCCD_DB_INIT should be set to the database details, since in this case this processor tries to get the parameters then from there. The string (either as parameter or as environment variable) should have the form "DB_HOST:DB_NAME:DB_USER:PASSWD". (string, optional, no default)
     * @param ConditionsTag This parameter sets the tag of the conditions data you are writing. If not set, no tagging will be performed (optional, string, default: "1.0")
     * @param WriteEveryNEvents This parameter gives the option to write the calculated pedestals every N events, if set to -1 the pedestals will be calculated and written after all events in the input file(s) have been read. (integer, default: -1)
     * @param ValidFrom This optional parameter sets the time the conditions data will be valid from in the conditions database. If not set, the time of the first event will be taken at the validuity start time. (lcio::long64)
     * @param ValidTill This optional parameter sets the time the conditions data will be valid till in the conditions database. If not set, the time of the last event will be taken at the validuity end time. (lcio::long64)
     * @param CalculationIterations How many times the calculation iterates over the raw data vector. (integer, default: 5)
     * @param IgnoreSurrounding # time slices before and after the currend slice must not be a signal. (integer, default: 5)
     * @param InitCalcWidth Start value for the pedestal width for the calculation. (float, default: 512.0)
     * @param Precision Precision of pedestal calculation. (double, default: 1.0E-5)
     * @param SignalThreshold Data time slice is considered a signal during calculation iteration, if is bigger than signalThreshold * pedestal width. (double, default: 5.0)
     * @param MakeRunNumberFolder If true, a sub-folder named like the run number will be created in the database folder DBFolder. The pedestals are then stored in this folder for this run (bool, default: false)
     * @author M. E. Janssen (DESY), R. Diener (DESY)
     */
    
    class PedestalCalculator : public marlin::Processor
    {

        public:
                        

            ///
            virtual Processor*  newProcessor() { return new PedestalCalculator ; }

            PedestalCalculator();

//            PedestalCalculator( const std::string &typeName ) : DBEntryMaker( typeName ){};

            /** Called at the begin of the job before anything is read.
             * Use to initialize the processor, e.g. book histograms.
             */
            virtual void init();

            /** Called for every run.
             */
            virtual void processRunHeader( lcio::LCRunHeader* run ) ;

            /** Called for every event - the working horse.
             */
            virtual void processEvent( lcio::LCEvent * evt ) ;

            virtual void check( lcio::LCEvent * evt ) ;

            /** Called after data processing for clean up.
             */
            virtual void end() ;

        protected:

            /// Structure to store pedestal value and width.
            struct StcPedestal
            {
                /// Pedestal Value.
                float Mean;
                /// Pedestal Width
                float Width;
            };
            
            /// Typedef: vector to store pedestal data.
            /** Typedef to store the pedestal (the StcPedestal struct contains the mean and the width of the pedestal) 
             *  in a vector 
             */
            typedef std::vector<StcPedestal*> PedestalVector_t;
            
            /// Structure to store the pedestal of a channel and other needed informations during internal calculations.
            struct ChannelPedestal
            {
                /// Hardware channel ID.
                int                 ChannelID;
                /// Hardware readout Group ID.
                int                 ReadoutGroup;
                /// Vector of raw ADC data.
                FloatVec*           Data;
                /// Vector to store calculated pedestal information in StcPedestal structs 
                /// (contains value and width).
                PedestalVector_t*   PedestalCol;
                /// Average of the PedestalCol vector.
                StcPedestal*        Average;
                /// Average error of the PedestalCol vector.
                StcPedestal*        AverageError;
            };    
            
                      
            /// Typedef: Map to store the pedestal data during calculation.
            /** Typedef of a map of ChannelPedestals to store the pedestal data during the calculation.
             * The key is a pair of integers, first entry: channel ID, second: readout group.
             * The ChannelPedestal struct contains: channel and readout group ID, 
             * a data vector (for ADC counts), a PedestalVector_t and tow structs for the average
             * of the pedestal value and width and the error of the average values
             */
            typedef std::map< std::pair<int, int>, ChannelPedestal* > PedestalMap_t;

            
            /// Map to store the pedestals internally during the calculation.
            PedestalMap_t*               _PedestalMap;
            
            /// Channel identifier, pair of two integers (first one for the pad and second one for the readout group)
            std::pair<int, int>          _ChannelPair;

            
            /// Claculates the pedestal mean and width from raw data.            
            /** Calculate the mean and the width of the pedestal out of the given raw data vector.
             *  @return A structure Pedestal containing these values (channel and readout group ID, 
             * a data vector [ADC counts], a PedestalVector_t and tow structs for the average
             * of the pedestal value and width and the error of the avergae values).
             */
            StcPedestal* calcPedestal(FloatVec* Data);

            
            /// Calculates average of pedestals of different events of a channel.
            /** Calculates the average of the values for mean and width of the pedestal
             *  of the given structure and save it in this structure.
             */
            virtual void calcAverage(ChannelPedestal* thisChannelPedestal);

            ///Creates an LCCollectionVector of Pedestals and fills it with the calculated pedestals. 
            virtual void       CreatePedestalCollection( PedestalMap_t*
                thisChannelPedestalCollection);

            /// Write the pedestal to the std output.
            virtual void                   printPedestal(PedestalMap_t*
                thisChannelPedestalCollection);

            /// Writes a summery of the pedestals to the std output.
            virtual void                   printPedestalShort(PedestalMap_t*
                thisChannelPedestalCollection);

            virtual void                   printPedestal(LCCollection* thisCollection);

            /// Converts the pedestal information to \link tpcconddata::Pedestal Pedestal objects \endlink and stores them into an LCCollection Vector
            virtual void makeLCCDData();

                        
            // Member variables for the processor parameters.
            /// Name of the input collection
            std::string _inputColName;
            /// Name of the pedestal collection
            std::string _pedestalColName;
            /// Name of the simple file where the pedestal collection should be stored (optional)
            std::string _simpleFileName;
            /// Name of the folder, where the pedestal collection should be stored in the database (optional)
            std::string _dbFolder;
            /// DBInit String (optional, can also be provided by evironment varibale)
            std::string _dbInitString;
            /// Tag of the conditions data that should be stored
            std::string _conditionsTag;
            /// Calculate and write the calculated pedestals after this number of events.
            int _writeEveryNEvents;
            /// Time from when the calculated pedestals will be valid (is stored in the database).
            std::string _validFrom;
            /// Time till when the calculated pedestals will be valid (is stored in the database).
            std::string _validTill;
            /// How many times the calculation iterates over the raw data vector.
            int _calculationIterations;
            /// # time slices before and after the currend slice must not be a signal.
            int _ignoreSurrounding;
            /// Start value for the pedestal width for the calculation.
            float  _initCalcWidth;
            /// Precision of pedestal calculation.
            double _precision;
            /// Data time slice is considered a signal during calculation iteration, if is bigger than signalThreshold * pedestal width.
            double _signalThreshold;
            /// parameter if sub-folder named by run number should be automatically added to the database folder
            bool _makeRunNumberFolder;
            
            // count the number of events
            int _NRun ;
            int _NEvt ;
            lcio::long64  _validFromInt;
            lcio::long64  _validTillInt;
            lcio::long64 _lastSinceTime;
            lcio::long64 _lastEventTime;            
            
        private:  
            
            void writePedestals(); 
            bool _doTagging;
            int _runNumber;
            int _writingIteration;
            std::string _detectorName;
            std::string _folderName;
            IMPL::LCCollectionVec* _outputCollection;
    };
}
#endif
