/* PedestalCalculator.cc
 * $Id: PedestalCalculator.cc,v 1.1 2006/07/07 08:58:52 mjanssen Exp $
 */
#include "PedestalCalculator.h"

//C++
#include <iostream>
#include <iomanip>
#include <map>
#include <cmath>
#include <utility>
#include <cstdio>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TPCHit.h>
#include <EVENT/TrackerData.h>
#include <EVENT/LCGenericObject.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCEventImpl.h>
#include <UTIL/LCTime.h>

//LCCD
#include <lccd.h>
#include <lccd/DBInterface.hh>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

    PedestalCalculator aPedestalCalculator;

    PedestalCalculator::PedestalCalculator() : Processor("PedestalCalculator")
    {
        // modify processor description
        _description = "PedestalCalculator: calculate the pedestal (mean and width) from raw data." ;

        registerProcessorParameter( "InputColName",
                                    std::string("Name for the input colletion. "),
                                    _inputColName,
                                    std::string("TPCConvertedRawData")
                                  );
        registerProcessorParameter( "PedestalColName",
                                    std::string("Name of the collection containing the pedestals. "),
                                    _pedestalColName,
                                    std::string("TPCPedestal")
                                  );
        registerOptionalParameter( "SimpleFileName" ,
                                   std::string("Base name of the simple LCIO file the pedestals should be written to. "),
                                   _simpleFileName,
                                   std::string("Pedestal") 
                                 );
        registerOptionalParameter( "DBFolder" ,
                                   std::string("Folder, where the pedestal collection should be stored in the database. "),
                                   _dbFolder,
                                   std::string("/test/pedestals") 
                                 );
        registerOptionalParameter( "DBInitString" ,
                                   std::string("Init spring for the connection to the database (optional). Format: DB_HOST:DB_NAME:DB_USER:PASSWD"),
                                   _dbInitString,
                                   std::string("localhost:lccd_test:calvin:hobbes")
                                 );        
        registerOptionalParameter( "ConditionsTag",
                                    std::string("Tag of the conditions data you are writing."),
                                    _conditionsTag,
                                    std::string("1.0")
                                  );
        registerProcessorParameter( "WriteEveryNEvents",
                                    std::string("Write the calculated pedestals every N events. If -1 pedestals will be written at the end of input."),
                                    _writeEveryNEvents,
                                    (int)-1
                                  );
        registerOptionalParameter( "ValidFrom" ,
                                   "Sets the time the conditions data will be valid from. If not set, valid from 1st event time. ",
                                   _validFrom,
                                   std::string("0")
                                 );
        registerOptionalParameter( "ValidTill" ,
                                   "Sets the time the conditions data will be valid till. If not set, valid till last event time. ",
                                   _validTill,
                                   std::string("4102444799000000000")
                                 );
        registerProcessorParameter( "CalculationIterations",
                                    std::string("How many times the calculation iterates over the raw data vector. "),
                                    _calculationIterations,
                                    (int)5
                                  );
        registerProcessorParameter( "IgnoreSurrounding",
                                    std::string("# time slices before and after the currend slice must not be a signal. "),
                                    _ignoreSurrounding,
                                    (int)5
                                  );
        registerProcessorParameter( "InitCalcWidth",
                                    std::string("Start value for the pedestal width for the calculation. "),
                                    _initCalcWidth,
                                    (float)512.0
                                  );
        registerProcessorParameter( "Precision",
                                    std::string("Precision of pedestal calculation. "),
                                    _precision,
                                    (double)1.0E-5
                                  );
        registerProcessorParameter( "SignalThreshold",
                                    std::string("Data time slice is considered a signal during calculation iteration, if is bigger than signalThreshold * pedestal width."),
                                    _signalThreshold,
                                    (double)5.0
                                  );
        registerProcessorParameter( "MakeRunNumberFolder",
                                    std::string("If true, a sub-folder named like the run number will be created in the database folder where are then stored for this run"),
                                    _makeRunNumberFolder,
                                    (bool)false);
    }

    
    void PedestalCalculator::init()
    {
        printParameters();
        _NRun = 0 ;
        _NEvt = 0 ;
        _lastSinceTime = -1;
        _lastEventTime = -1;
        _writingIteration = 0;
        std::istringstream streamFrom(_validFrom);
        std::istringstream streamTill(_validTill);
        streamFrom >> _validFromInt;
        streamTill >> _validTillInt;
        _PedestalMap = new PedestalMap_t();
        _PedestalMap->clear();
        _doTagging=false;
    }

    
    void PedestalCalculator::processRunHeader( LCRunHeader* run)
    {
        m_out(DEBUG)<<"PedestalCalculator::processRunHeader"<<std::endl;
        run->parameters().setValue( _processorName + "_revision", "$Rev: 1818 $");
        for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
        {
            if ( ! i->second->isOptional() || i->second->valueSet() )
            {
                run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
            }
        }
        if (_NRun > 0) 
        {
            _doTagging=true;            
            makeLCCDData();
            writePedestals();            
            m_out(MESSAGE)<<std::endl<<"Wrote pedestal collection for run "<<_runNumber<<" (Collection #"<<std::setfill('0') << std::setw(3) << _writingIteration<<" for this run)"<<std::endl;
            _lastSinceTime = -1;
            _writingIteration = 0;
        }
        _runNumber    = run->getRunNumber();
        _detectorName = run->getDetectorName();
        ++_NRun ;
    }

    
    void PedestalCalculator::processEvent( LCEvent * evt )
    {
        //read input data

        LCCollection* InputCollection = evt->getCollection( _inputColName ) ;

        if (_NEvt==0)
        {
            //first event
            if ( _lastSinceTime < 0 )
            {
                //or the time of the first event
                _lastSinceTime = evt->getTimeStamp();
            }
            
        }
        if ( _lastSinceTime < 0 ) _lastSinceTime = evt->getTimeStamp();
        _lastEventTime = evt->getTimeStamp();        
        _NEvt+=1;
        
        if( InputCollection != 0 )
        {
            int nPulse = InputCollection->getNumberOfElements();
            FloatVec* Data = new FloatVec();
            for(int iPulse=0; iPulse< nPulse ; iPulse++)
            {
                TrackerData* thisPulse = dynamic_cast<TrackerData*>( InputCollection->getElementAt( iPulse ) ) ;
                if (thisPulse)
                {
                    _ChannelPair = std::make_pair(thisPulse->getCellID0(),thisPulse->getCellID1()) ;
                    ChannelPedestal* thisChannelPedestal = 0;
                    PedestalMap_t::iterator iChannelPedestal = 
                            _PedestalMap->find(_ChannelPair);
                    if ( iChannelPedestal != _PedestalMap->end() ) 
                    {
                            thisChannelPedestal = (*iChannelPedestal).second;
                    }
                    else
                    {
                        // pad does not exist
                        m_out(DEBUG)<<"create new ChannelPedestal"<<std::endl;
                        thisChannelPedestal = new ChannelPedestal();
                        thisChannelPedestal->ChannelID   = thisPulse->getCellID0();
                        thisChannelPedestal->ReadoutGroup    = thisPulse->getCellID1();
                        thisChannelPedestal->PedestalCol = new PedestalVector_t();
                        _PedestalMap->insert(std::make_pair(_ChannelPair, thisChannelPedestal) );
                    }
                    // clear the data from data of older pulses
                    ///std::cout<<"clear data"<<std::endl;
                    Data->clear();
                    // fill in new data
                    FloatVec TrackerDataSpectrum = thisPulse->getChargeValues();
                    m_out(DEBUG)<<"fill data"<<std::endl;
                    for (FloatVec::iterator iSlice=TrackerDataSpectrum.begin();
                        iSlice!=TrackerDataSpectrum.end(); iSlice++)
                    {
                        // fill the data vector with
                        Data->push_back( *iSlice );
                    }
                    m_out(DEBUG)<<"calculate pedestal"<<std::endl;
                    thisChannelPedestal->PedestalCol->push_back( calcPedestal( Data ) );
                }
            }
            delete Data;
        }
        // give alive message once in a while
        if (_NEvt%100 == 0)
        {
            m_out(MESSAGE)<<"Number of events processed: "<<std::setfill('0')<<std::setw(6)<<_NEvt<<std::endl;
        }
        if ( parameterSet("WriteEveryNEvents") && ( (_NEvt%_writeEveryNEvents) == 0 ) )       
        {
            m_out(MESSAGE)<<"EveryNEvents is reached for the "<< _writingIteration<<". time: writing the collection."<<std::endl;
            makeLCCDData();
            writePedestals();
            _lastSinceTime = -1;
            ++_writingIteration;
            m_out(MESSAGE)<<std::endl<<"Writing of pedestal collection for run "<<_runNumber<<" (Collection #"<<std::setfill('0') << std::setw(3) << _writingIteration<<" for this run) is finished."<<std::endl;
        }
    }

    
    void PedestalCalculator::check( LCEvent * evt )
    {
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }

    
    void PedestalCalculator::end()
    {
        _doTagging=true;
        makeLCCDData();
        writePedestals();
        m_out(MESSAGE)<<std::endl<<"Wrote pedestal collection for run "<<_runNumber<<" (Collection #"<<std::setfill('0') << std::setw(3) << _writingIteration<<" for this run)"<<std::endl;
        m_out(MESSAGE)<<"End of the data input: Cleaning up."<<std::endl;
        delete _PedestalMap;
    }

    
    PedestalCalculator::StcPedestal* PedestalCalculator::calcPedestal(FloatVec* Data)
    {
        // initialize the calculation parameters/variables
        float calcValue  = 0.0;
        float calcWidth  = _initCalcWidth;
        int   cycleNorm  = 0;
        float cycleValue = 0.0;
        float cycleWidth = 0.0;

                
        for (int iRun=0; iRun < _calculationIterations; iRun++)
        {
            m_out(DEBUG)<<"cycle "<<iRun<<std::endl;
            cycleNorm  = 0;
            cycleValue = 0.0;
            cycleWidth = 0.0;
            FloatVec::iterator iData = Data->begin();

            // initialization, which finds iData slice without signal (inside _ignoreSurrounding region): 
            // makes sure, that pedestal calculation starts in area with no signal 
            int NoSignalCounter = 0;
            while  ( ( iData != Data->end() ) && ( NoSignalCounter < 2*_ignoreSurrounding+1 ) )
            {
                if ( ( (*iData)-calcValue <  _signalThreshold*calcWidth) &&
                     ( (*iData)-calcValue > -_signalThreshold*calcWidth) )
                {
                    // is no signal
                    NoSignalCounter++;
                }
                else
                {
                    // is signal
                    NoSignalCounter = 0;
                }
                iData++;
            } // end initialization loop: while  ( ( iData != Data->end() ) && ( NoSignalCounter < 2*_ignoreSurrounding+1 ) )
            
            iData-=_ignoreSurrounding;

            // loop over areas without signal
            while ( (iData+_ignoreSurrounding) < Data->end() )
            {
                if ( ( (*(iData+_ignoreSurrounding))-calcValue <  _signalThreshold*calcWidth ) &&
                     ( (*(iData+_ignoreSurrounding))-calcValue > -_signalThreshold*calcWidth ) )
                {
                    // no signal
                    cycleNorm++ ;
                    cycleValue += (*iData);
                    cycleWidth += (*iData)*(*iData);
                    iData++;
                }
                else
                {
                    //signal: find next region without signal
                    NoSignalCounter = 0;
                    while  ( (iData!=Data->end()) && (NoSignalCounter < 2*_ignoreSurrounding+1) )
                    {
                        if ( ( (*iData)-calcValue <  _signalThreshold*calcWidth) &&
                             ( (*iData)-calcValue > -_signalThreshold*calcWidth) )
                        {
                            // is no signal
                            NoSignalCounter++;
                        }
                        else
                        {
                            // is signal
                            NoSignalCounter = 0;
                        }
                        iData++;
                    }
                    iData-=_ignoreSurrounding;
                }  // end else of if statement: 
                   //if ( ( (*iData+_ignoreSurrounding)-calcValue <  _signalThreshold*calcWidth ) &&
                   //     ( (*iData+_ignoreSurrounding)-calcValue > -_signalThreshold*calcWidth ) )
            } // end "while" loop over areas without signal: while ( (iData+_ignoreSurrounding) < Data->end() )
            
            if (cycleNorm>0)
            {
                if (fabs(calcValue-cycleValue/cycleNorm) < _precision) break;
                calcValue = cycleValue/cycleNorm;
                calcWidth = sqrt(cycleWidth/cycleNorm - calcValue*calcValue);
                // s = 1/n*sum(x-<x>)^2=<x^2> - <x>^2
            }
            
        } // end "for" loop iterations: for (int iRun=0; iRun < _calculationIterations; iRun++)
        
        // return calculated values
        StcPedestal* returnPedestal= new StcPedestal();
        returnPedestal->Mean  = calcValue;
        returnPedestal->Width = calcWidth;
        return returnPedestal;
    }

    
    void PedestalCalculator::calcAverage(ChannelPedestal* thisChannelPedestal)
    {
        StcPedestal*  Average      = new StcPedestal();
        StcPedestal*  AverageError = new StcPedestal();

        // calculate value and its error, for pedestal and width
        Average->Mean       = 0.0;
        AverageError->Mean  = 0.0;
        Average->Width      = 0.0;
        AverageError->Width = 0.0;
        if ( thisChannelPedestal->PedestalCol->size() > 0)
        {
            // loop over the internal vector and sum up the values of the elements
            for (PedestalVector_t::iterator iChannelPedestal = thisChannelPedestal->PedestalCol->begin();
                iChannelPedestal != thisChannelPedestal->PedestalCol->end(); iChannelPedestal++)
            {
                Average->Mean       += (  *iChannelPedestal )->Mean;
                AverageError->Mean  += ( (*iChannelPedestal )->Mean  ) * ( (*iChannelPedestal)->Mean );
                Average->Width      += (  *iChannelPedestal )->Width;
                AverageError->Width += ( (*iChannelPedestal )->Width ) * ( (*iChannelPedestal)->Width );
            }
            // calculate the means
            Average->Mean  /= thisChannelPedestal->PedestalCol->size();
            Average->Width /= thisChannelPedestal->PedestalCol->size();
            // calculate the errors
            if ( thisChannelPedestal->PedestalCol->size() > 1 )
            {
                AverageError->Mean = sqrt( (AverageError->Mean / thisChannelPedestal->PedestalCol->size()
                    - (Average->Mean * Average->Mean ) )
                    / ( thisChannelPedestal->PedestalCol->size()-1) );
                AverageError->Width = sqrt( (AverageError->Width / thisChannelPedestal->PedestalCol->size()
                    - (Average->Width * Average->Width ) )
                    / ( thisChannelPedestal->PedestalCol->size()-1) );
                // s^2 = <x^2> - <x>^2 => sigma = s / sqrt(n-1)
            }
            else
            {
                AverageError->Mean  = Average->Mean;
                AverageError->Width = Average->Width;
            }
        }
        else
        {
            // no values to calculate
            Average->Mean       = 0.0;
            Average->Width      = 0.0;
            AverageError->Mean  = 0.0;
            AverageError->Width = 0.0;
        }
        // set the values in the referenced object
        thisChannelPedestal->Average      = Average;
        thisChannelPedestal->AverageError = AverageError;
    }

    void PedestalCalculator::CreatePedestalCollection( PedestalMap_t*
        thisPedestalMap )
    {
        // fill the pedestal map in one LCCollectionVector, 
        // structure has to be resorted in the PedestalHandler when reading from the database
        _outputCollection = new LCCollectionVec( LCIO::LCGENERICOBJECT );        
        for (PedestalMap_t::iterator iChannelPedestal = thisPedestalMap->begin();
             iChannelPedestal != thisPedestalMap->end(); iChannelPedestal++)
        {
            calcAverage( (iChannelPedestal->second) );
            _outputCollection->addElement( new tpcconddata::Pedestal( 
                                            (iChannelPedestal->second)->ChannelID,
                                            (iChannelPedestal->second)->ReadoutGroup,
                                            (iChannelPedestal->second)->Average->Mean,
                                            (iChannelPedestal->second)->Average->Width ) 
                                        );
        }
    }

    
    void PedestalCalculator::makeLCCDData()
    {
	m_out(DEBUG4)<<"This is PedestalCalculator::makeLCCDData(), size of pedestal map is "
		     <<_PedestalMap->size() << std::endl;

	m_out(DEBUG4)<<"store pedestals: valid "<<_lastSinceTime<<" - "<<_lastEventTime<<std::endl;
 	

        if ( (_PedestalMap->size()>0)&&( _lastSinceTime<_lastEventTime ) )
        {
            m_out(MESSAGE)<<std::endl<<std::endl;
            m_out(MESSAGE)<<"store pedestals: valid "<<_lastSinceTime<<" - "<<_lastEventTime<<std::endl;
            printPedestalShort( _PedestalMap );
            CreatePedestalCollection( _PedestalMap );
            _outputCollection->parameters().setValue("Revision",tpcconddata::Pedestal::getRevision() );
            //clean up of pedestal collection
            for (PedestalMap_t::iterator iChannelPedestal = _PedestalMap->begin();
                iChannelPedestal != _PedestalMap->end(); iChannelPedestal++)
            {
                (iChannelPedestal->second)->PedestalCol->clear();
                delete ( (iChannelPedestal->second)->Average );
                delete ( (iChannelPedestal->second)->AverageError );
            }
        }
    }

    
    void PedestalCalculator::printPedestal(PedestalMap_t* thisPedestalMap)
    {
        // output
        m_out(MESSAGE)<<std::endl;
        for ( PedestalMap_t::iterator iChannelPedestal = thisPedestalMap->begin();
            iChannelPedestal != thisPedestalMap->end(); iChannelPedestal++ )
        {
            calcAverage( iChannelPedestal->second );
            printf( "Channel/Module: %5.0i/%5.0i  ",(iChannelPedestal->second)->ChannelID,(iChannelPedestal->second)->ReadoutGroup );
            printf( "Pedestal      : mean = %6.4f+-%6.4f",
                    (iChannelPedestal->second)->Average->Mean, (iChannelPedestal->second)->AverageError->Mean);
            if ( (iChannelPedestal->second)->Average->Width == 0 )
            {
                m_out(WARNING)<<" width calculation failed";
            }
            else
            {
                printf("   width = %6.4f+-%6.4f",
                       (iChannelPedestal->second)->Average->Width, (iChannelPedestal->second)->AverageError->Width);
            }
            m_out(MESSAGE)<<std::endl;
        }
    }

    
    void PedestalCalculator::printPedestalShort(PedestalMap_t*
        thisPedestalMap)
    {
        // output
        ChannelPedestal*  firstChannelPedestal = ((thisPedestalMap->begin())->second);
        calcAverage( firstChannelPedestal );
        std::pair<int,int> firstChannel(firstChannelPedestal->ChannelID,firstChannelPedestal->ReadoutGroup);
        std::pair<int,int> lastChannel                  = firstChannel;
        std::pair<int,int> smallestPedestalMeanChannel  = firstChannel;
        std::pair<int,int> largestPedestalMeanChannel   = firstChannel;
        std::pair<int,int> smallestPedestalWidthChannel = firstChannel;
        std::pair<int,int> largestPedestalWidthChannel  = firstChannel;
        float smallestPedestalMean  = firstChannelPedestal->Average->Mean;
        float largestPedestalMean   = smallestPedestalMean;
        float smallestPedestalWidth = firstChannelPedestal->Average->Width;
        float largestPedestalWidth  = smallestPedestalWidth;
        bool  noWidthCalFailed = true;

        for (PedestalMap_t::iterator iChannelPedestal = thisPedestalMap->begin();
            iChannelPedestal != thisPedestalMap->end(); iChannelPedestal++)
        {
            calcAverage( iChannelPedestal->second );
            std::pair<int,int> Channel((iChannelPedestal->second)->ChannelID,(iChannelPedestal->second)->ReadoutGroup);
            if (Channel<firstChannel) firstChannel = Channel;
            if (Channel>lastChannel)  lastChannel  = Channel;
            float PedestalMean = (iChannelPedestal->second)->Average->Mean;
            if (PedestalMean<smallestPedestalMean)
            {
                smallestPedestalMean = PedestalMean;
                smallestPedestalMeanChannel = Channel;
            }
            if (PedestalMean>largestPedestalMean)
            {
                largestPedestalMean  = PedestalMean;
                largestPedestalMeanChannel = Channel;
            }
            float PedestalWidth = (iChannelPedestal->second)->Average->Width;
            if ( PedestalWidth == 0 )
            {
                if (noWidthCalFailed)
                {
                    m_out(MESSAGE)<<"width calculation failed for channel: ";
                    noWidthCalFailed=false;
                }
                m_out(MESSAGE)<<Channel.first<<"/"<<Channel.second<<", ";
            }
            else
            {
                if (PedestalWidth<smallestPedestalWidth)
                {
                    smallestPedestalWidth = PedestalWidth;
                    smallestPedestalWidthChannel = Channel;
                }
                if (PedestalWidth>largestPedestalWidth)
                {
                    largestPedestalWidth  = PedestalWidth;
                    largestPedestalWidthChannel = Channel;
                }
            }
        }
        if (!noWidthCalFailed)
        {
            m_out(MESSAGE)<<std::endl;
        }
        m_out(MESSAGE)<<"calculate pedestals for channel/module "<<firstChannel.first<<"/"<<firstChannel.second<<
                " to "<<lastChannel.first<<"/"<<lastChannel.second<<std::endl;
        m_out(MESSAGE)<<"pedestal mean between "
                <<smallestPedestalMean<<
                " (channel/module "<<smallestPedestalMeanChannel.first<<"/"<<smallestPedestalMeanChannel.second
                <<") and "
                <<largestPedestalMean<<" (channel/module "<<largestPedestalMeanChannel.first<<"/"<<
                largestPedestalMeanChannel.second<<")"<<std::endl;
        m_out(MESSAGE)<<"pedestal width between "
                <<smallestPedestalWidth<<" (channel/module "<<smallestPedestalWidthChannel.first
                <<"/"<<smallestPedestalWidthChannel.second
                <<") and "
                <<largestPedestalWidth<<" (channel/module "<<largestPedestalWidthChannel.first
                <<"/"<<largestPedestalWidthChannel.second<<")"<<std::endl;
    }

    
    void PedestalCalculator::printPedestal(LCCollection* thisCollection)
    {
        // output
        int         nPedestal    = thisCollection->getNumberOfElements();
        tpcconddata::Pedestal* thisPedestal = 0;
        m_out(MESSAGE)<<std::endl;
        for (int iPedestal = 0; iPedestal<nPedestal; iPedestal++)
        {
            thisPedestal = new tpcconddata::Pedestal( dynamic_cast<LCGenericObject*>
                ( thisCollection->getElementAt( iPedestal ) ) );
            printf( "Channel/Module: %5.0i/%5.0i  ",
                (thisPedestal->getHardwareID()).first,(thisPedestal->getHardwareID()).second );
            printf( "Pedestal: mean = %6.4f", thisPedestal->getPedestalValue() );
            if ( thisPedestal->getPedestalWidth() == 0 )
            {
                m_out(WARNING)<<" width calculation failed";
            }
            else
            {
                printf( "   width = %6.4f", thisPedestal->getPedestalWidth() );
            }
            m_out(MESSAGE)<<std::endl;
        }
    }
    
    
    void PedestalCalculator::writePedestals()
    {
        if ( parameterSet("DBFolder")  )
        {
            if ( _makeRunNumberFolder )
            {
                std::stringstream out;
                out << "/" << _runNumber;
                _folderName=_dbFolder+out.str();
                m_out(MESSAGE)<<"Changed database folder from: "<<_dbFolder<<" to "<<_folderName<<std::endl;
            }
            else
            {
                _folderName=_dbFolder;
            }
            lccd::DBInterface* dbConnection;
            if ( parameterSet("DBInitString") ) 
            {
                dbConnection = new lccd::DBInterface(_dbInitString, _folderName, true);
            } else {
                dbConnection = new lccd::DBInterface(_folderName, true);
            }
            lcio::long64 from;
            lcio::long64 till;
            if ( parameterSet("ValidFrom") ) 
            {
                from=_validFromInt;
            } else {
                from=_lastSinceTime;
            }
            if ( parameterSet("ValidTill") ) 
            {
                till=_validTillInt;
            } else {
                till=_lastEventTime;
            }
            dbConnection->storeCollection( from, till, _outputCollection, "TPCPedestal");
            if ( parameterSet("ConditionsTag") )
            {
                if ( _makeRunNumberFolder ) 
                {
                    try
                    {
                        dbConnection->tagFolder( _conditionsTag, std::string("TPCPedestal Tag ") + _conditionsTag + std::string(" by MarlinTPC pedestalcalculator") );
                        m_out(MESSAGE)<<"Folder "<<_folderName<<" has been tagged to tag "<<_conditionsTag<<" ."<<std::endl;
                    }
                    catch( ... )
                    {
                        m_out(WARNING)<<"Error while tagging the folder  "<<_folderName<<" (most probably the tag already existed): folder is NOT tagged."<<std::endl;
                    }
                }
            } 
            _doTagging=false;
            delete dbConnection;
            if ( !parameterSet("SimpleFileName") ) delete _outputCollection;
        }
        
        if ( parameterSet("SimpleFileName") ) 
        {
            // create a LCWriter
            IO::LCWriter* lcWrt = lcio::LCFactory::getInstance()->createLCWriter();
            std::string fileName;
            std::stringstream out;
            out << _runNumber;
            if ( parameterSet("WriteEveryNEvents")  ) out << std::string("_") << std::setfill('0') << std::setw(3) << _writingIteration;
            fileName = _simpleFileName + std::string("_") + out.str() + std::string(".slcio");
            
            try 
            {
                lcWrt->open(fileName);
                LCRunHeaderImpl* runHdr = new LCRunHeaderImpl(); 
                runHdr->setRunNumber( _runNumber ) ;
                runHdr->setDetectorName( _detectorName ) ;
                runHdr->setDescription( std::string("Pedestals calculated with MarlinTPC pedestal calculator.")  ) ;
                runHdr->parameters().setValue( _processorName + "_revision", "$Rev: 1818 $");
                for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
                {
                    if ( ! i->second->isOptional() || i->second->valueSet() )
                    {
                        runHdr->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
                    }
                }
                lcWrt->writeRunHeader( runHdr );
                LCEventImpl*  evt = new LCEventImpl();
                evt->setRunNumber( _runNumber  );
                evt->setEventNumber( 0 );
                evt->setTimeStamp( _lastSinceTime ) ;  
                evt->setDetectorName( _detectorName ) ;
                evt->addCollection( _outputCollection , "TPCPedestals" ) ;
                lcWrt->writeEvent( evt ) ;
            
                try
                {
                    lcWrt->close();
                } 
                catch (IOException)
                {
                    m_out(ERROR)<<"Error closing the simple file. Status of data in "<<fileName<<" unknown."<<std::endl;
                }
                delete evt;
                delete runHdr;
                delete lcWrt;
            } 
            catch (IOException)
            {
                m_out(ERROR)<<"Error opening the simple file: File "<<fileName<<" already exists. Collection is NOT stored."<<std::endl;
            }
        }
        if ( !parameterSet("DBFolder") && !parameterSet("SimpleFileName") )
        {
            m_out(ERROR)<<"Error: No parameters set where to write the calculated pedestals. Collection is NOT stored"<<std::endl;
        }
        _outputCollection=0;
    }
    
    
}
