#ifdef USE_LCCD

// MarlinTPC
#include "PhotoelectricFittingProcessor.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearxml/tinyxml.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

// Minuit
#include <Minuit2/MnPrint.h>
#include <Minuit2/MnUserParameters.h>
#include <Minuit2/MnUserCovariance.h>
#include <Minuit2/MnMinimize.h>
#include <Minuit2/FunctionMinimum.h>

using namespace CLHEP;
using namespace EVENT;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace ROOT;
using namespace Minuit2;
using namespace std;

namespace marlintpc {


PhotoelectricFittingProcessor aPhotoelectricFittingProcessor;


PhotoelectricFittingProcessor::PhotoelectricFittingProcessor() : Processor("PhotoelectricFittingProcessor"),

	_input_photoelectric_tracker_pulses_collection_name("TPCPhotoelectricPulses"),
	_output_photodot_collection_name("TPCPhotodots"),
	_output_photoline_collection_name("TPCPhotolines"),
	_output_is_transient(true),
	_projected_photogeometry_file_name("projected_photogeometry.xml"),
	_sigma(2.5) {

	/* description of the processor which will be displayed in the
	* steering file autometicly generated by
	* marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	*/

	_description = "PhotoelectricFittingProcessor performs a likelihood fit on photoelectric data.";

	/* register steering parameters: name, description, class-variable, default value
	* the type will by definded by the type of the default value
	* string, double, float and int are possible
	* plaese note to get a float you must cast in C++: (float) 1.3
	*/

	registerInputCollection( LCIO::TRACKERHIT ,
		"InputPhotoelectricTrackerPulses",
		"The name of the input collection of TPCTrackerPulses from the photoelectric system.",
		_input_photoelectric_tracker_pulses_collection_name,
		std::string("TPCPhotoelectricPulses") );

	registerOutputCollection( LCIO::TRACKERHIT ,
		"OutputPhotodots",
		"The name of the output collection of TPCHits which represent fitted photodots.",
		_output_photodot_collection_name,
		std::string("TPCPhotodots") );

	registerOutputCollection( LCIO::TRACKERHIT ,
		"OutputPhotolines",
		"The name of the output collection of TPCHits which represent fitted photolines.",
		_output_photoline_collection_name,
		std::string("TPCPhotolines") );

	registerOptionalParameter( "SetOutputTransient",
		"Controls whether the output collection should be saved in the LCIO file (false) or if it should be deleted at the end (true - default).",
		_output_is_transient,
		true);

	registerOptionalParameter( "ProjectedPhotogeometryFilename",
		"This is the name of the collection which stores the true positions of the photodots as SimTrackerHit objects. It is optional but it provides the number of photodots and, if provided, the processor will order the seed photodots to match the order of true photodots (useful for analysis).",
		_projected_photogeometry_file_name,
		_projected_photogeometry_file_name );


	registerOptionalParameter( "Sigma",
		"The diffusion of the dot.",
		_sigma,
		_sigma );

}


void PhotoelectricFittingProcessor::init() {

	// usually a good idea to
	printParameters();

	// load AIDA (if it's available)
	AIDA::ITuple *projected_photodot_tuple = 0;
	try {

		_photodot_fit_results_tuple = AIDAProcessor::tupleFactory( this )->create( "PhotodotFitResults", "Photodot Fit Results", "int i_dot, float x, float y", "" );
		projected_photodot_tuple = AIDAProcessor::tupleFactory( this )->create( "PhotodotTuple", "Projected Photodot Stats", "int i_dot, int pad_count", "" );

	}
	catch ( lcio::Exception &e ) {

		streamlog_out( DEBUG ) << e.what() << endl;
		return;
	}

	// open the projection file and get the time interval we're watching for and the list of pads to take
	TiXmlDocument document ( _projected_photogeometry_file_name.c_str() );

	if( !document.LoadFile() ) {

		streamlog_out( ERROR ) << "Couldn't load projected geometry file " << _projected_photogeometry_file_name << "." << endl;
		return;
	}

	int i_dot = 0;

	// get a list of global pad indices to search for
	TiXmlNode *photodot_node = document.RootElement()->FirstChild("photodot");

	for ( ; photodot_node; photodot_node = photodot_node->NextSibling( "photodot" ) ) {

		// create a set to hold the pulses for this photodot
		std::list< std::set< EVENT::TrackerPulse* > >::iterator new_pulse_set = _photodot_pulse_sets.insert( _photodot_pulse_sets.end(), set< EVENT::TrackerPulse* >() );

		TiXmlElement *pad_element = photodot_node->FirstChildElement( "pad" );

		int pad_count = 0;

		// add the pads
		for ( ; pad_element; pad_element = pad_element->NextSiblingElement( "pad" ) ) {

			int pad_index;
			int module_id;

			pad_element->QueryIntAttribute( "index", &pad_index );
			pad_element->QueryIntAttribute( "module", &module_id );

			GlobalPadIndex new_index( pad_index, module_id );

			// assign this global index to this pulse collection
			_photodot_pulse_set_by_global_index[ new_index ] = new_pulse_set;

			pad_count += 1;
		}

		if ( projected_photodot_tuple ) {

			projected_photodot_tuple->fill( 0, i_dot );
			projected_photodot_tuple->fill( 1, pad_count );
			projected_photodot_tuple->addRow();
		}

		// and the centroid
		TiXmlElement *centroid_element = photodot_node->FirstChildElement( "centroid" );

		if ( centroid_element ) {

			string centroid_text = centroid_element->FirstChild()->Value();

			size_t comma_pos = centroid_text.find (',');

			string coord_text = centroid_text.substr ( 1, comma_pos - 1 );
			const float x = atof ( coord_text.c_str() );

			coord_text = centroid_text.substr ( comma_pos + 1, centroid_text.length() - comma_pos - 1 );
			const float y = atof ( coord_text.c_str() );

			_photodot_centroids.push_back ( Hep2Vector( x, y ) );

			streamlog_out( DEBUG ) << "photodot " << i_dot << " has centroid " << centroid_text << "." << endl;
		}

		i_dot++;
	}

	TiXmlNode *photoline_node = document.RootElement()->FirstChild("photoline");

	for ( ; photoline_node; photoline_node = photoline_node->NextSibling( "photoline" ) ) {

		TiXmlElement *pad_element = photoline_node->FirstChildElement( "pad" );

		for ( ; pad_element; pad_element = pad_element->NextSiblingElement( "pad" ) ) {

			int pad_index;
			int module_id;

			pad_element->QueryIntAttribute( "index", &pad_index );
			pad_element->QueryIntAttribute( "module", &module_id );

			streamlog_out( DEBUG ) << "Valid pad index " << pad_index << " module " << module_id << "." << endl;
		}
	}

}


void PhotoelectricFittingProcessor::processRunHeader( LCRunHeader* run) {

}


void PhotoelectricFittingProcessor::processEvent( LCEvent * evt ) {

	// Step 1. Reset the pulse data because the TrackerPulses will be different from the last event
	for ( std::list< std::set< EVENT::TrackerPulse* > >::iterator pulse_set = _photodot_pulse_sets.begin(); pulse_set != _photodot_pulse_sets.end(); pulse_set++ ) {

		pulse_set->clear();
	}
	
	// Step 2. Associate the TrackerPulses with the correct photodot or photoline
	LCCollection* input_photoelectric_pulses = evt->getCollection( _input_photoelectric_tracker_pulses_collection_name );

	streamlog_out( MESSAGE ) << "Using " << input_photoelectric_pulses->getNumberOfElements() << " pulses from the collection " << _input_photoelectric_tracker_pulses_collection_name << " to fit " << _photodot_pulse_sets.size() << " photodots." << endl;

	// now determine if this is a photoelectric pulse
	for ( int i_tracker_pulse = 0; i_tracker_pulse < input_photoelectric_pulses->getNumberOfElements(); i_tracker_pulse++ ) {

		TrackerPulse* input_pulse = dynamic_cast<TrackerPulse*>( input_photoelectric_pulses->getElementAt( i_tracker_pulse ) ) ;

		if ( !input_pulse ) {

			throw( SkipEventException( this ) );
		}

		const GlobalPadIndex global_index( input_pulse->getCellID0(), input_pulse->getCellID1() );

		// I can't wait for the new auto keyword...
		const std::map< gear::GlobalPadIndex, std::list< std::set< EVENT::TrackerPulse* > >::iterator >::iterator i_pulse_set = _photodot_pulse_set_by_global_index.find ( global_index );

		if ( i_pulse_set != _photodot_pulse_set_by_global_index.end() ) {

			i_pulse_set->second->insert( input_pulse );
		}
	}

	// Step 3. Loop through the photodots, fitting them
	int i_dot = 0;
	list< Hep2Vector >::const_iterator centroid = _photodot_centroids.begin();

	for ( list< set< TrackerPulse* > >::const_iterator i_photodot = _photodot_pulse_sets.begin(); i_photodot != _photodot_pulse_sets.end(); i_photodot++, centroid++, i_dot++ ) {

		streamlog_out( DEBUG ) << "Fitting photodot " << i_dot << " with centroid " << *centroid << "." << endl;

		PhotodotLikelihood photodot_likelihood( *i_photodot );

		MnUserParameters user_parameters;

		user_parameters.Add( "x", (*centroid)[0], 1.E-01 );
		user_parameters.Add( "y", (*centroid)[1], 1.E-01 );
		user_parameters.Add( "sigma", _sigma );

		MnMinimize minimizer( photodot_likelihood, user_parameters );

		FunctionMinimum minimum = minimizer();

		if ( !minimum.IsValid() ) {

			continue;
		}

		streamlog_out ( DEBUG ) << minimum;

		const MnUserParameters &fit_parameters = minimum.UserParameters();

		_photodot_fit_results_tuple->fill( 0, i_dot );
		_photodot_fit_results_tuple->fill( 1, fit_parameters.Value( "x" ) );
		_photodot_fit_results_tuple->fill( 2, fit_parameters.Value( "y" ) );
		_photodot_fit_results_tuple->addRow();
	}
}



void PhotoelectricFittingProcessor::check( LCEvent * evt ) {

	// nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PhotoelectricFittingProcessor::end() {

}


PhotoelectricFittingProcessor::PhotodotLikelihood::PhotodotLikelihood( const set< TrackerPulse* > &photodot_data ):
			_photodot_data( photodot_data ) {

}


double PhotoelectricFittingProcessor::PhotodotLikelihood::operator()( const vector<double>& x ) const {

	// the parameters are global x, global y, and sigma
	const double sigma = x[2];

	// get the TPC parameters
	const gear::TPCParameters &tpc_parameters = Global::GEAR->getTPCParameters();

	double negative_log_likelihood = 0.;

	float c0, c1;
	// convert the parameters to the global coordinate system of the TPC
	if ( tpc_parameters.getCoordinateType() == PadRowLayout2D::CARTESIAN ) {

		c0 = x[0];
		c1 = x[1];
	}
	else if ( tpc_parameters.getCoordinateType() == PadRowLayout2D::POLAR ) {

		c0 = sqrt( x[0]*x[0] + x[1]*x[1] );
		c1 = atan2( x[1], x[0] );
	}

	// iterate through the pads calculating how much charge they were expected to collect

	for( set< TrackerPulse* >::const_iterator i_data_pulse = _photodot_data.begin(); i_data_pulse != _photodot_data.end(); i_data_pulse++ ) {

		const TrackerPulse *pulse = *i_data_pulse;

		const TPCModule &module = tpc_parameters.getModule( pulse->getCellID1() );

		Vector2D local_photodot_coords = module.globalToLocal( c0, c1 );

		Vector2D global_pad_centre = module.getPadCenter( pulse->getCellID0() );

		Vector2D local_pad_centre = module.globalToLocal( global_pad_centre[0], global_pad_centre[1] );

		double a;
		double b;

		double half_pad_width = 0.5 * module.getPadWidth( pulse->getCellID0() );
		double half_pad_height = 0.5 * module.getPadHeight( pulse->getCellID0() );

		// convert the result to cartesian coordinates
		if ( module.getCoordinateType() != PadRowLayout2D::CARTESIAN ) {

			if ( module.getCoordinateType() == PadRowLayout2D::POLAR ) {

				float r = local_photodot_coords[0];
				local_photodot_coords[0] = r * cos( local_photodot_coords[1] );
				local_photodot_coords[1] = r * sin( local_photodot_coords[1] );

				r = local_pad_centre[0];
				half_pad_width *= r;
				const double sin_p = sin( local_pad_centre[1] );
				const double cos_p = cos( local_pad_centre[1] );

				local_pad_centre[0] = r * cos_p;
				local_pad_centre[1] = r * sin_p;

				// rotate the displacement vector so the pad height is aligned with the local y-axis
				const double ap = local_pad_centre[0] - local_photodot_coords[0];
				const double bp = local_pad_centre[1] - local_photodot_coords[1];

				a = (ap * sin_p) - (bp * cos_p);
				b = (bp * sin_p) + (ap * cos_p);
			}
			else {

				return 9999999.;
			}


		} else {

			a = local_pad_centre[0] - local_photodot_coords[0];
			b = local_pad_centre[1] - local_photodot_coords[1];
		}

		// now perform the integral
		const double d = M_SQRT1_2 / sigma;
		const double A = erf( (a - half_pad_width)*d  ) - erf( (a + half_pad_width)*d  );
		const double B = erf( (b - half_pad_height)*d ) - erf( (b + half_pad_height)*d );
		const double expected_charge = A*B;

		negative_log_likelihood -= std::log( expected_charge ) * pulse->getCharge();
	}

	return negative_log_likelihood;
}


double PhotoelectricFittingProcessor::PhotodotLikelihood::Up() const {

	return 0.5;
}


}

#endif // USE_LCCD
