#ifndef CHANNELCORRECTIONHANDLER_H
#define CHANNELCORRECTIONHANDLER_H

// tpcconddata
#include <ChannelCorrection.h>

// stl
#include <map>

// lccd
#include "lccd/IConditionsChangeListener.hh"

namespace marlintpc {

/** 
 *  A helper class for convenient usage of the channel correction information.
 *  It allows to obtain information about calibration constants, time offsets,
 *  broken or noisy channel, etc.
 *
 *  @author Peter Wienemann, University of Bonn
 */
class ChannelCorrectionHandler: public lccd::IConditionsChangeListener {

public:

	 /**
	* Constructor for use with lccd
	*
	* @param collection_name the name of the collection registered with lccd in the conditions processor
	*/
	ChannelCorrectionHandler( std::string collection_name );
    
	/**
	* Get channel quality
	*
	* @param channelID hardware channel ID
	*
	* @return integer with channel quality bits
	*/
	int getChannelQuality( int channelID );

	/**
	* Get number of calibration factors
	* (there can be an arbitrary number of calibration factors per channel)
	*
	* @param channelID hardware channel ID
	*
	* @return number of calibration factors
	*/
	int getNChannelCalibrationFactors( int channelID );

	/**
	* Get calibration factor
	* (there can be an arbitrary number of calibration factors per channel)
	*
	* @param channelID hardware channel ID
	* @param index index of calibration factor
	*
	* @return channel calibration factor
	*/
	float getChannelCalibrationFactor( int channelID, int index = 0 );

	/**
	* Get calibration factor vector
	* (there can be an arbitrary number of calibration factors per channel)
	*
	* @param channelID hardware channel ID
	*
	* @return vector with calibration factors
	*/
	std::vector<float> getChannelCalibrationFactors( int channelID );

	/**
	* Get channel time offset
	*
	* @param channelID hardware channel ID
	*
	* @return time offset in ns
	*/
	float getChannelTimeOffset( int channelID );

	/**
	* Get information about channel operability
	*
	* @param channelID hardware channel ID
	*
	* @return true if channel is broken
	*/
	bool isBroken( int channelID );

	/**
	* Get information about noise in channel
	*
	* @param channelID hardware channel ID
	*
	* @return true if channel is noisy
	*/
	bool isNoisy( int channelID );

	/**
	* Called by LCCD whenever the conditions change
	*
	*/
	void conditionsChanged( EVENT::LCCollection* col );

 private:

	 typedef std::map<int, tpcconddata::ChannelCorrection> channel_map_t;

	 channel_map_t _channel_corrections_by_channel;

	/**
	* Get a reference to a channel correction object if it exists in the map
	*
	* @param channelID hardware channel ID
	*/
	tpcconddata::ChannelCorrection& getChannelCorrection( int channelID );

};

} // namespace marlintpc

#endif // CHANNELCORRECTIONHANDLER_H
