#ifndef CHANNEL_MAPPER_H
#define CHANNEL_MAPPER_H

#include <marlin/Processor.h>
#include <lcio.h>
#include <string>
#include <EVENT/TrackerPulse.h>
#include "MappingHandler.h"

namespace marlintpc
{

/** The ChannelMapperProcessor changes cellID0 to map from hardware channel numbers
 *  to sowftware/logical channel numbers.
 *
 *  @author Martin Killenberg, University of Bonn
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of TrackerPulses
 *  \li A collection with the mapping (ADCChannelMapping)
 *
 *  <h4>Output</h4>
 *  \li A collection of TrackerPulses. If no output collection name is given,
 *      the input collection is modified.
 *
 *  @param InputCollectionName Name of input data collection (default: TPCPulses)
 *  @param MappingCollectionName Name of the collection with the mapping (default: TPCChannelMapping)
 *  @param OutputCollectionName \c optional Name of output data collection
 *  @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 *   Only applies if the \c OutputCollectionName is given and a new collection is created.
 *   If the input collection is modified, the status from the creating processor is preserved.
 *
 *  \exception lcio::ReadOnlyException
 *  Thrown in case the input collection is read only and no output collection name
 *  is given
 *  \exception marlin::SkipEventException
 *  Thrown if input collection not available or empty
 */

class ChannelMapperProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new ChannelMapperProcessor ;
        }


        ChannelMapperProcessor() ;

        /** Write information into the run header
         */
        virtual void processRunHeader(EVENT::LCRunHeader* run);

        /** Called for every event - the working horse.
         */
        virtual void processEvent(EVENT::LCEvent * evt) ;


        /** Called after data processing for clean up.
         */
        virtual void end() ;


    protected:

    private:

        std::string _inputColName ;    ///< the name of the input collection
        std::string _outputColName ;   ///< the name of the output collection
        std::string _mappingColName ;   ///< the name of the collection with the mapping data

/// set the persistency flag of the output collection
        bool _outputIsPersistent;

        MappingHandler myMappingHandler; ///< instance of the mapping handler
} ;

}//namespace marlintpc

#endif//CHANNEL_MAPPER_H

