#ifndef PedestalHandler_h
#define PedestalHandler_h 1

//C++
#include <iostream>
#include <utility>
#include <stdexcept>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//TPCCondData
#include "Pedestal.h"

// the package schould be capsulated in a name space
namespace marlintpc
{

    //#ifdef USE_LCCD
    /** Helper Class which provides access to the pedestals for every event.
     *  It is derived from IConditionsChangeListener,
     *  so the pedestals are updated every time the conditions data change.
     *  (Obsolete? One could use LCCDConditionsMap instead)
     * @author M. E. Janssen (DESY), R. Diener (DESY)
     */

    class PedestalHandler : public lccd::IConditionsChangeListener
    {
        //#else
        //    class PedestalHandler {
        //#endif

        public:

            // Several Typedefs used in following declarations
            /// Integer pair of Hardware IDs, first entry: channelID, second: readoutGroup;
                                 // ( channel ID, module ID )
            typedef std::pair<int, int> ModuleChannelPair_t;
            /// Map of channel pedestals.
            typedef std::map< int, tpcconddata::Pedestal* > PedestalPadMap_t;
            /// Typedef: Map (for readout groups) of map (for channels) of Pedestals
            typedef std::map< int, PedestalPadMap_t* >      PedestalModuleMap_t;

            //#ifdef USE_LCCD
            /** Constructor
             * @param ColName name of the collection in which the pedestals are stored
             */
            PedestalHandler( std::string ColName = tpcconddata::Pedestal::getDefaultColName() );
            //#else
            //    /** constructor (without use of LCCD)
            //     * @param ColName name of the collection in which the pedestals are stored
            //     * @param FileName name of the file in wwhich the pedestals are stored
            //     */
            //    PedestalHandler( std::string ColName, std::string FileName );
            //#endif
            /// Destructor
            virtual ~PedestalHandler( );

            static std::string getDefaultColName()
            {
                return  tpcconddata::Pedestal::getDefaultColName();
            };

            /// Updates stored pedestal values if conditions have change.
            /** Called by the Marlin ConditionsProcessor, if the condition change
             *  inplemetentation of the IConditionsChangeListener interface
             */
            void conditionsChanged( EVENT::LCCollection* col );

            /// Get Pedestal Object for a channel.
            /** Returns the stored \link tpcconddata::Pedestal Pedestal object \endlink for the given channel
             *  @param hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the pad
             *  @return \link tpcconddata::Pedestal Pedestal object \endlink
             */
            tpcconddata::Pedestal* getPedestal( ModuleChannelPair_t hardwareID );

            /// Get Pedestal Object for a channel.
            /** Returns the stored \link tpcconddata::Pedestal Pedestal object \endlink for the given channel.
             *  @param channelID,readoutGroup channel ID (cellID0) and readout group (cellID1) of the pad
             *  @return \link tpcconddata::Pedestal Pedestal object \endlink
             */
            tpcconddata::Pedestal* getPedestal( int channelID, int readoutGroup ) ;

            /// Get pedestal value for a channel.
            /** Returns the pedestal value for the given channel.
             *  @param hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the pad
             *  @return Pedestal mean value (0.0 if no value is stored for the ID).
             */
            float getPedestalValue( ModuleChannelPair_t hardwareID );

            /// Get pedestal value for a channel.
            /** Returns the pedestal value for the given channel.
             *  @param channelID,readoutGroup Channel ID (cellID0) and readout group (cellID1) of the pad
             *  @return Pedestal mean value (0.0 if no value is stored for the ID).
             */
            float getPedestalValue( int channelID, int readoutGroup );

            /// Get pedestal width for a channel.
            /** Returns the pedestal width for the given channel.
             *  @param hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the pad
             *  @return Pedestal width (-1.0 if no value is stored).
             */
            float getPedestalWidth( ModuleChannelPair_t hardwareID );

            /// Get pedestal width for a channel.
            /** Returns the pedestal width for the given channel.
             *  @param channelID,readoutGroup Channel ID (cellID0) and readout group (cellID1) of the pad
             *  @return Pedestal width (-1.0 if no value is stored).
             */
            float getPedestalWidth( int channelID, int readoutGroup );

        private:
            /// Map to store the pedestals of the channels of one readout group.
            PedestalPadMap_t    _pedestalPadMap;
            /// Map to store the maps of the channels of the readout groups.
            PedestalModuleMap_t _pedestalModuleMap;
    };

}
#endif
