#ifndef TimePixClusterProjectionSeparatorProcessor_h
#define TimePixClusterProjectionSeparatorProcessor_h 1

// C++
#include <string>
#include <map>
#include <vector>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "TimePixMapListener.h"
#include "lccd/ConditionsMap.hh"


//AIDA
//#include <AIDA/AIDA.h>

//ROOT
#include "TFile.h"
#include "THStack.h"

namespace AIDA{

  class IHistogram1D;
  class IHistogram2D;

}

namespace marlintpc{

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  
     *  converts TrackerHits of a TimePix into further separated TrackerHits, by projecting them onto their own clusteraxis.
     *  This projection is weighted by charge information if available, that is if the pixel was operated in TOT mode.\n
     *  Conditionsdata must be stored in a seperate collection of TimePixPixelMode for every event in which the mode of operation changes.\n
     *  Handling of conditions data is done by the ConditionsProcessor
     *
     *  <h4>Input - Prerequisites</h4>
     *  a collection of TrackerHits containing clusters of adjacent pixels
     *
     *  <h4>Output</h4> 
     * @li Collection containing TrackerHits of further separated clusters, separation by projection along cluster axis.
     * @li Collection of tracker data objects which has been split because it belonged to several clusters
     * 
     * @param InputCollectionName Name of the input collection with unseparated hit candidates (default "TimePixHitCandidates")
     * @param OutputCollectionName Name of the output containing with sepatated hit candidates (default "TimePixSepHitCandidates")
     * @param OutputTrackerDataCollectionName Name of the output collection with sepatated tracker data objects (default "TimePixSepData")
     * @param OutputRelationsCollectionName Name of the collection with relations between clusters and tracker data (default "TimePixRelationSepHit")
     * @param nColumn number of columns on the chip
     * @param nRow number of rows on the chip
     * @param clusterSizeToSep number of pixels in a cluster for it to be separated (default 9)
     * @param SetOutputTransient If not 0 the output collection is set transient
     * 
     * @author Simone Zimmermann
     */

    class TimePixClusterProjectionSeparatorProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here TimePixClusterProjectionSeparatorProcessor)
	 */
	virtual Processor*  newProcessor() { return new TimePixClusterProjectionSeparatorProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	TimePixClusterProjectionSeparatorProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;

	/* Finds the next local maximum in an AIDA::IHistogram1D,
	 * starting at the binnumber given by int startbin
	 * and returns the bin number of the maximum 
	 * if there is no maximum to be found it returns 0
	*/
	virtual int locMaxFinder( int startbin, AIDA::IHistogram1D * h1) ;

	/* Finds the next local minimum in an AIDA::IHistogram1D,
	 * starting at the binnumber given by int startbin
	 * and returns the bin number of the minimum 
	 * if there is no minimum to be found it returns 0
	*/
	virtual int locMinFinder( int startbin, AIDA::IHistogram1D * h1) ;

	/* Finds the bin at which the cluster should be separated
	 * starting at the local minimum given by the binnumber locmin
	 * if the cluster should not be separated it returns 0
	*/
	virtual int sepBinFinder( int locmin, AIDA::IHistogram1D * h1) ;

    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
	std::string _outputColName ; ///<- the name of the output tracker hits collection
	std::string _outputTrackerDataColName ; ///<- the name of the output tracker data collection
	std::string _relationsColName ; ///<- the name of the colletion with relations between hits and tracker data

	int _nColumn; ///<- number of columns on the chip
	int _nRow; ///<- number of rows on the chip
	int _clusterSizeToSep; ///<- number of pixels in a cluster in order for it to be separated

	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

	/** is true till the first output collection is writen to an event
	 *  used to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	bool isFirstCollection;

	int _nRun ; ///<- counts the processed runs

	double a; ///<- parameter of axis of projection y = a*x + b
	double b; ///<- parameter of axis of projection y = a*x + b
	
	/** the actual mode map is a lccd::ConditionsMap
	 *  the key is a PixelKey, specifically a struct which contains
	 *  the ChipID and the PixelID
	 *  the elements are TimePixPixelModes
	 */
	tpcconddata::TimePixMap * _timePixMap;

    private:
	std::vector<AIDA::IHistogram1D *> hProjectXVec;
//	AIDA::IHistogram1D * highTOT;
	AIDA::IHistogram2D * aidahistocluster;
	AIDA::IHistogram2D ** aidaonehit;


    } ;
    
}

#endif
    
    

