#ifndef TimePixHitCenterCalculatorProcessor_h
#define TimePixHitCenterCalculatorProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "lccd/ConditionsMap.hh"
#include "TimePixMapListener.h"

namespace AIDA {
  class IHistogram2D;
  class IHistogram3D;

}

namespace marlintpc{

   /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  
     *  calculates the center of gravity of each cluster, weighted by charge information.
     *
     *  <h4>Input - Prerequisites</h4>
     *  a collection of TrackerHits, position not set
     *
     *  <h4>Output</h4> 
     *  Collection of TrackerHits with the position and dEdx set.
     * 
     * @param InputCollectionName Name of input collection, TrackerHits without position
     * @param OutputCollectionName Name of output collection, TrackerHits with position
     * @param ModeMapCollectionName Name of mode map collection
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param nColumn number of columns on the chip
     * @param nRow number of rows on the chip
     * @param pitch pitch of the pixel
     * @param SetOutputTransient If not 0 the output collection is set transient
     * 
     * @author Simone Zimmermann
     */


    class TimePixHitCenterCalculatorProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here TimePixHitCenterCalculatorProcessor)
	 */
	virtual Processor*  newProcessor() { return new TimePixHitCenterCalculatorProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	TimePixHitCenterCalculatorProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;
	
    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
	std::string _outputColName ; ///<- the name of the output collection
	std::string _modeMapColName ; ///<- the name of the mode map collection
	std::string _modeMapCorrName ;///<- the name of the correction to mode map collection


	double _shutter; ///<- length of shutter signal in us
	double _readoutFrequency; ///<- readout frequency of TimePix chip in MHz, optional, should come from gear
	double _driftVelocity; ///<- drift velocity of gas in mm/us
	bool _multiShutter; ///<- true if multiple shutters are allowed during acquisition

	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

	tpcconddata::TimePixMap * _timePixMap;


    private:
//	AIDA::IHistogram2D * histoCluster;
//	AIDA::IHistogram2D * histoSP;	
//	AIDA::IHistogram3D * histo3D;	

    } ;
    
}

#endif
    
    

