#ifndef TimePixMultiChipBoardDividerProcessor_h
#define TimePixMultiChipBoardDividerProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"

namespace AIDA{
  class IHistogram2D;
}

namespace marlintpc{

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  TimePixMultiChipBoardDividerProcessor of the MarlinTPC package .
     * 
     *   converts MultiTrackerRawData of multi-TimePix_Chips into TrackerRawData of each TimePix
     *
     *  <h4>Input - Prerequisites</h4>
     *  a collection of MultiTrackerRawData
     *
     *  <h4>Output</h4> 
     *  Collection containing TrackerRawData.
     * 
     * @param InputCollectionName Name of input collection, TrackerRawData
     * @param OutputCollectionName Name of output collection, TrackerRawData
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param PixelsPerChipX number of columns of the chip
     * @param PixelsPerChipY number of rows of the chip
     * @param ChipsPerBoardX  number of Chips in a row
     * @param ChipsPerBoardY number of Chips  in a column
     * 
     * @author Frederik Kloeckner
     */
    class TimePixMultiChipBoardDividerProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here TimePixMultiChipBoardDividerProcessor)
	 */
	virtual Processor*  newProcessor() { return new TimePixMultiChipBoardDividerProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	TimePixMultiChipBoardDividerProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;
	
	
    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
	std::string _outputColName ; ///<- the name of the output collection
	int _pixelsPerChipX ; ///<-Number of columns of the chip
	int _pixelsPerChipY ; ///<-Number of rows of the chip
	int _chipsPerBoardY ; ///<-Number of chips in a column
	int _chipsPerBoardX   ; ///<-Number of chips in a row
	
	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

	/** is true till the first output collection is wirten to an event
	 *  used to to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	//bool isFirstCollection;

	//int _nRun ; ///<- counts the processed runs


private:
	AIDA::IHistogram2D* histoCluster;
	AIDA::IHistogram2D* histoClusterRaw;
	
    } ;
    
}

#endif
    
    

