#ifndef TRACKFITTERLIKELIHOODPROCESSOR_H
#define TRACKFITTERLIKELIHOODPROCESSOR_H 1

#include "marlin/Processor.h"

#include "TrackFitterBase.h"

#include "TPCConditionsHandler.h"

// LCIO
#include "lcio.h"

/* NOTE this uses the C++ version of Minuit - not the Fortran version */
#include "Minuit2/FCNBase.h"

// stl
#include <string>
#include <vector>

// some forward declarations

namespace AIDA {
	
	class ITuple;
}

namespace EVENT {
	class LCEvent;
	class LCRunHeader;
	class Track;
}

namespace IMPL {

	class TrackImpl;
	class TrackerHitImpl;
}

namespace marlintpc {

/** TrackFitterLikelihoodProcessor : This class takes post-amplified data from a readout pad and
 * reconstructs the track based on the location of each signal. The algorithm was developed by Dean Karlen at UVic
 *    @author J. Abernathy (UVic)<br>
 */
class TrackFitterLikelihoodProcessor : public marlin::Processor, public ROOT::Minuit2::FCNBase {

public:

	/** Default constructor
	*/
	TrackFitterLikelihoodProcessor();


	/** Destructor
	*/
	~TrackFitterLikelihoodProcessor();


	/** Return a new instance of this processor
	*/
	marlin::Processor * newProcessor();


	/** Called at the begin of the job before anything is read.
	*/
	void init();


	/** Called once per run to process the event's header
	*/
	void processRunHeader (EVENT::LCRunHeader *run);


	/** Called for every event - the working horse.
	*/
	virtual void processEvent(EVENT::LCEvent *evt);


	// From FCNBase
	/**
		The meaning of the vector of parameters is of course defined by the user,
		who uses the values of those parameters to calculate his function value.
		The order and the position of these parameters is strictly the one specified
		by the user when supplying the starting values for minimization. The starting
		values must be specified by the user, either via an std::vector<double> or the
		MnUserParameters supplied as input to the MINUIT minimizers such as
		VariableMetricMinimizer or MnMigrad. Later values are determined by MINUIT
		as it searches for the minimum or performs whatever analysis is requested by
		the user.

		@param par function parameters as defined by the user.

		@return the value of the function.

		@see MnUserParameters
		@see VariableMetricMinimizer
		@see MnMigrad
	*/
	virtual double operator()(const std::vector<double>& x) const;

	/**
	Error definition of the function. MINUIT defines parameter errors as the
	change in parameter value required to change the function value by up. Normally,
	for chisquared fits it is 1, and for negative log likelihood, its value is 0.5.
	If the user wants instead the 2-sigma errors for chisquared fits, it becomes 4,
	as Chi2(x+n*sigma) = Chi2(x) + n*n.
	\todo Comment a little bit better with links!!!!!!!!!!!!!!!!! Idem for errorDef()
	*/
	virtual double Up() const;


	/** Called for every event - right after processEvent() has been called for all processors.
	*/
	virtual void check (lcio::LCEvent *evt);


	/** Called after data processing for clean up in the inverse order of the init() method so that resources allocated in the first processor also will be available for all following processors.
	*/
	virtual void end();


private:

	/** Takes the track parameters in the expected track and calculates the expected charge distribution on the readout electronics
	 */
	void calculate_expected_charge_distribution( TrackImpl &expected_track, double sigma0 ) const;

	/** A helper function for calculate_expected_distribution, calculates the expected distrubition due to a uniform/homogenous field
	*/
	void calculate_expected_charge_distribution_homogeneous_polar(IMPL::TrackImpl&, const double) const;

	/** A helper function for calculate_expected_distribution, calculates the expected distrubition due to a uniform/homogenous field
	*/
	void calculate_expected_charge_distribution_homogeneous_cartesian(IMPL::TrackImpl&, const double) const;

	/** A helper function for calculate_expected_distribution, calculates the expected distrubution due to a non-homogenous field
	*/
	void calculate_expected_charge_distribution_non_homogeneous_polar(IMPL::TrackImpl&, const double) const;

	/** A helper function for calculate_expected_distribution, calculates the expected distribution due to a non-homogenous field
	*/
	void calculate_expected_charge_distribution_non_homogeneous_cartesian(IMPL::TrackImpl&, const double) const;

	/** A helper function for calculate_expected_distribution, calculates the expected charge distribution on a polar row
	*/
	void calculate_expected_charge_distribution_polar_row(IMPL::TrackerHitImpl&, const EVENT::TrackerHit &data_hit, const double, const double, const double ) const;

	/** A helper function for calculate_expected_distribution, calculates the expected charge distribution on a cartesian row
	*/
	void calculate_expected_charge_distribution_cartesian_row(IMPL::TrackerHitImpl&, const EVENT::TrackerHit &data_hit, const double, const double, const double) const;

	/** Calculates the likelihood of observing this distribution of charge on the readout pads in the readout geometry plane (r-phi/x-y)
	*/
	const double calculate_charge_distribution_log_likelihood(const EVENT::Track&) const;


	/** Calculates the likelihood of observing this distribution of primary ionizations in the z direction with a specified z deviation
	*/
	const double calculate_primary_ionization_location_log_likelihood( const Track& ) const;


	/** Uses MINUIT to find the most likely track which fits the data supplied in the seed track
	*/
	IMPL::TrackImpl find_most_likely_fit(EVENT::Track* seed_track);


	/** Calculates the expected distribution of charge on one pad by the line-gaussian parameterized using Dean Karlen's calculations. see "TPC Performance in Magnetic Fields with GEM and Pad Readout"
	* b is the distance from the centre of the pad to the track in the x dimension (negative is left) (mm)
	* phi is the local azimuthal angle of the straight line segment in the pad row (relative to [0,1,0], clockwise is negative)
	* sigma is the standard deviation of the distribution
	* height is the height of the pad (mm)
	* width is the width of the pad (mm)
	*/
	double calculate_expected_distribution(double b, double phi, double sigma,
		double height, double width) const;

	/** The name of the LCCollection to obtain the seed track from.
	*/
	std::string _seed_track_collection_name;

	/** The name of the LCCollection to store the fitted track into the event with.
	*/
	std::string _fitted_track_collection_name;

	/** the name of the collection the TPC conditions are stored with
	*/
	std::string _tpc_conditions_collection_name;

	/** The constant diffusion of the GEMs and similar TPC structures which does not vary depending on distance traveled
	*/
	double _sigma0;

	/** Should _sigma_0 be a free parameter in the minimization?
	* true : free parameter
	* false : constant (default)
	*/
	bool _fit_sigma0;

	/** The transverse diffusion constant of the gas in the TPC in sqrt(mm)
	*/
	double _gas_trans_diffusion_constant;

	/** The transverse defocussing constant of the gas in the TPC in mm
	*/
	double _gas_trans_defocussing_constant;

	/** A noise factor to reduce the affect of spurious signals
	*/
	double _p_noise;

	/** The track being minimized currently
	*/
	EVENT::Track* _current_seed_track;

	/** The innermost position of the readout electronics
	**/
	double _readout_electronics_inner_z;

	/** The standard deviation of the primary ionizations in the z-direction
	*/
	double _sigma_z;

	/** a parameter which chooses whether the track tuples are produced
	 */
	bool _produce_tuples;

	/** the track parameters stored as tuples
	*/
	AIDA::ITuple *_track_result_tuples;

	/** the gas parameterization of the drift volume of the TPC
	 */
	double _omega_tau;

	/** the parameterized drift velocity in the drift section of the TPC (mm/us)
	 */
	double _drift_velocity;

	/** the path integral distance (mm) between points of the expected track for non-homogenous magnetic field tracking
	 * FIXME: this is a constant value right now. it could be more accurate by adding extra points where the magnetic field is strongest
	 */
	double _path_integral_step_size;

	/** The outer radius of the TPC (mm)
	 */
	double _tpc_outer_radius;

	/** determines whether the processor will attempt to correct for non-homogenous magnetic fields (slows down fitting)
	 */
	bool _fit_nonhomogeneous_fields;

	/** scaling parameters to keep the five track parameters on the same order of magnitude
	 */
	int _omega_order_of_magnitude;
	int _tanLambda_order_of_magnitude;

	/** the helper class for determining TPC conditions during runtime
	*/
	TPCConditionsHandler *_tpc_conditions;

	/** A helper class for the likelihood processor. It encapsulates a curved gaussian distribution of charge produced by a track. The polar fitting routine uses it when calculating the charge deposited on a pad by the track
	*/
	/*class PolarPDF {

		public:
			PolarPDF (const Track &estimated_track);

			double evaluate_pdf_integral( const double r, const double phi );

		private:
			const double _estimated_omega;
			const double _estimated_phi0;
			const double _estimated_d0;
	}*/
};

} // namespace marlintpc

#endif // TRACKFITTERLIKELIHOODPROCESSOR_H
