#include "ADCElectronicsListener.h"

#include <utility>
#include <map>
#include <iostream>
#include <string.h>
#include <exception>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd_exceptions.h>

//lcio
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "UTIL/LCFixedObject.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

    ADCElectronicsListener::ADCElectronicsListener(std::string ConditionsName, std::string ColName )
    {
        //initialize first the members, else it goes wrong
        _genericADCElectronicsParametersMap.clear();
        _conditionsName=ConditionsName;
        _collectionName=ColName;
        _genericADCElectronicsParametersMap.clear();
        _thisEmptyCollection = new lcio::LCCollectionVec( LCIO::LCGENERICOBJECT );
        _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionADCElectronicsListener" );
        _timeStampSince=0;
        _timeStampTill=0;
        _useLastValidIfNoDataAvailable=false;
        _noDataAvailable=true;
        _condHandler=NULL;
        // now register this listener with the conditions processor
        if ( ! (marlin::ConditionsProcessor::registerChangeListener( this, ColName ) ) )
        {
            std::stringstream message;
            message << " ADCElectronicsListener couldn't register collection " << ColName << " with the ConditionsProcessor.";
            throw lccd::LCCDException(message.str());
        }
    }



    ADCElectronicsListener::~ADCElectronicsListener()
    {
        delete _thisEmptyCollection;
    }



    void ADCElectronicsListener::conditionsChanged( lcio::LCCollection* col )
    {
        m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged()"<<std::endl;

        // get collection time parameters and set internal members
        std::stringstream sstrSince(col->getParameters().getStringVal( lccd::DBSINCE ));
        sstrSince >> _timeStampSince;
        lcio::LCTime timeSince( _timeStampSince ) ;
        std::stringstream sstrTill(col->getParameters().getStringVal( lccd::DBTILL ));
        sstrTill >> _timeStampTill;
        lcio::LCTime timeTill( _timeStampTill ) ;
        m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): _timeStampSince="<<_timeStampSince<<" _timeStampTill="<<_timeStampTill<<std::endl;

        // look into the map to see if we have accepted this collection as a default
        std::map<lcio::LCCollection* ,lccd::IConditionsHandler* >::iterator itDefaultCollectionMap = _handlerDefaultCollectionMap.find(col);

        // check if the collection is our default collection
        if ( itDefaultCollectionMap != _handlerDefaultCollectionMap.end())
        {
            _noDataAvailable=true;

            m_out(MESSAGE)<<"ADCElectronicsListener::conditionsChanged(): default collection "<<col->getParameters().getStringVal( "CollectionName" )<<" sent "<<std::endl;
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): address of default collection: "<<col<<std::endl;

            //give some information on the missing conditions data
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): CollectionName: "<<_collectionName<<" ConditionsType: "<<_conditionsName<<" not available at this timestamp."<<std::endl;
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): No data in database from: "<<timeSince.timeStamp()<<" ("<<timeSince.year()<<" "<<timeSince.month()<<" "<<timeSince.day()<<" "<<timeSince.hour()<<" "<<timeSince.min()<<" "<<timeSince.sec()<<" "<<timeSince.ns()<<")"<<std::endl;
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): No data in database till: "<<timeTill.timeStamp()<<" ("<<timeTill.year()<<" "<<timeTill.month()<<" "<<timeTill.day()<<" "<<timeTill.hour()<<" "<<timeTill.min()<<" "<<timeTill.sec()<<" "<<timeTill.ns()<<")"<<std::endl;

            //check what to do if there are not data available
            if (_useLastValidIfNoDataAvailable)
            {
                if( _condHandler->lastValidCollection()!=NULL )
                {
                    fillConditionsMap(_condHandler->lastValidCollection());
                }
                else
                {
                    // no last valid collection available, so we bail out by throwing an exception
                    std::stringstream out;
                    out << "ADCElectronicsListener::conditionsChanged(): Tried to use last valid collection but found none. Exiting processing";
                    throw lccd::DataNotAvailableException(out.str());
                }
            }
            else
            {
                // no conditions data available, so we bail out by throwing an exception
                std::stringstream out;
                out << "ADCElectronicsListener::conditionsChanged(): No conditions data stored for current time stamp. Exiting processing";
                throw lccd::DataNotAvailableException(out.str());
            }
        }
        else                     // it is not a default so we can do the updating
        {
            _noDataAvailable=false;

            //give some information on the current conditions data
            m_out(MESSAGE)<<"ADCElectronicsListener::conditionsChanged(): new values available. Updating parameter map."<<std::endl;
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): CollectionName: "<<col->getParameters().getStringVal( "CollectionName" )<<std::endl;
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): Valid Since: "<<col->getParameters().getStringVal( "DBSince" )<<std::endl;
            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): Valid Till: "<<col->getParameters().getStringVal( "DBTill" )<<std::endl;

            fillConditionsMap(col);
        }
    }



    void ADCElectronicsListener::fillConditionsMap( lcio::LCCollection* col )
    {
        _genericADCElectronicsParametersMap.clear();
        int nEntries = col->getNumberOfElements();

        if (_conditionsName == "GenericADCElectronicsParameters" )
        {
            _genericADCElectronicsParametersMap.clear();
            tpcconddata::GenericADCElectronicsParameters* Entry;                std::pair<GenericADCElectronicsParametersMap_t::iterator,bool> iInsertReturn;

            m_out(DEBUG)<<"ADCElectronicsListener::conditionsChanged(): update conditions data  (data for "<<nEntries<<" modules in collection)"<<std::endl;

            for (int iEntry = 0; iEntry < nEntries; ++iEntry)
            {
                Entry = new tpcconddata::GenericADCElectronicsParameters(
                    dynamic_cast<LCGenericObject*>( col->getElementAt( iEntry ) ));
                iInsertReturn = _genericADCElectronicsParametersMap.insert(std::make_pair(Entry->getReadoutGroup(), Entry));
                if ( !iInsertReturn.second )
                {
                    m_out(ERROR)<<"ADCElectronicsListener: could not store GenericADCElectronicsParameters for readout group: ("<<Entry->getReadoutGroup()<<") in the conditions map because there was already a value stored for this readout group number."<<std::endl;
                }
            }                    //loop over all entries in the LCCollectionVec
        }
        else
        {
            throw tpcconddata::CondDataWrongObjectException(std::string("Conditions Object " + _conditionsName + " not implemented in Listener."));
        }
    }



    int ADCElectronicsListener::getPolarity(int readoutGroupID)
    {
        if (_conditionsName == "GenericADCElectronicsParameters" )
        {
            GenericADCElectronicsParametersMap_t::iterator Entry = _genericADCElectronicsParametersMap.find(readoutGroupID);
            if ( Entry == _genericADCElectronicsParametersMap.end() )
            {
                std::stringstream out;
                out << "ADCElectronicsListener: Conditions data not available for readout group id "<<readoutGroupID<<".";
                throw tpcconddata::CondDataNotAvailableException(out.str());
            }
            else
            {
                return (Entry->second)->getPolarity();
            }
        }
        else
        {
            std::stringstream out;
            out << "ADCElectronicsListener: Conditions data Polarity not available for " + _conditionsName + " object.";
            throw tpcconddata::CondDataNotAvailableException(out.str());
        }
    }



    int ADCElectronicsListener::getMaxADCValue(int readoutGroupID)
    {
        if (_conditionsName == "GenericADCElectronicsParameters" )
        {
            GenericADCElectronicsParametersMap_t::iterator Entry = _genericADCElectronicsParametersMap.find(readoutGroupID);
            if ( Entry == _genericADCElectronicsParametersMap.end() )
            {
                std::stringstream out;
                out << "ADCElectronicsListener: Conditions data not available for readout group id "<<readoutGroupID<<".";
                throw tpcconddata::CondDataNotAvailableException(out.str());
            }
            else
            {
                return (Entry->second)->getMaxADCValue();
            }
        }
        else
        {
            std::stringstream out;
            out << "ADCElectronicsListener: Conditions data MaxADCValue not available for " + _conditionsName + " object.";
            throw tpcconddata::CondDataNotAvailableException(out.str());
        }
    }



    float ADCElectronicsListener::getFrequency(int readoutGroupID)
    {
        if (_conditionsName == "GenericADCElectronicsParameters" )
        {
            GenericADCElectronicsParametersMap_t::iterator Entry = _genericADCElectronicsParametersMap.find(readoutGroupID);
            if ( Entry == _genericADCElectronicsParametersMap.end() )
            {
                std::stringstream out;
                out << "ADCElectronicsListener: Conditions data not available for readout group id "<<readoutGroupID<<".";
                throw tpcconddata::CondDataNotAvailableException(out.str());
            }
            else
            {
                return (Entry->second)->getFrequency();
            }
        }
        else
        {
            std::stringstream out;
            out << "ADCElectronicsListener: Conditions data Frequency not available for " + _conditionsName + " object.";
            throw tpcconddata::CondDataNotAvailableException(out.str());
        }
    }




    void ADCElectronicsListener::registeredWithHandler( lccd::IConditionsHandler* condHandler )
    {

        _condHandler=condHandler;
        m_out(MESSAGE)<<"ADCElectronicsListener::registeredWithHandler(): registered with:"<<condHandler->name()<<" adress: "<<condHandler<<std::endl;
        m_out(DEBUG)<<"ADCElectronicsListener::registeredWithHandler(): register default collection"<<std::endl;

        // try to get the default collection
        try
        {
            LCCollection* col = condHandler->defaultCollection();

            if( ! col )          // it will be null if none has so far been registered. So let's register ours
            {
                condHandler->registerDefaultCollection( _thisEmptyCollection );
                m_out(DEBUG)<<"ADCElectronicsListener::registeredWithHandler(): default collection registered; address: "<<_thisEmptyCollection<<std::endl;
                // and put in the map for this handler
                _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
            }
            //then the default handler was already registered, that's odd ... ;)
            else if( col == _thisEmptyCollection )
            {
                m_out(DEBUG)<<"ADCElectronicsListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:"<<std::endl;
            }
            else                 // somebody has got there before us, let's see if we like the default
            {
                // if it has the same CollectionName, we can use it, no problem ...
                if(col->getParameters().getStringVal("CollectionName") == "EmptyCollectionADCElectronicsListener")
                {
                    m_out(DEBUG)<<"ADCElectronicsListener::registeredWithHandler(): Default Collection already registered."<<std::endl;
                    _handlerDefaultCollectionMap[col] = condHandler;
                }
                // if not, we throw an exception because things should be handled in a uniform standardized way
                else
                {
                    m_out(ERROR)<< "ADCElectronicsListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
                    throw lccd::InconsistencyException("ADCElectronicsListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
                }
            }
        }
        // catch MemberNotImplemented exception, which happens, if a simple file is used as input
        catch (lccd::MemberNotImplementedException)
        {
            m_out(MESSAGE)<<"ADCElectronicsListener::registeredWithHandler: Could not register default collection. Seems simple file input is used for ADCelectronics conditions."<<std::endl;
        }
    }



    void ADCElectronicsListener::deRegisteredWithHandler( lccd::IConditionsHandler* condHandler )
    {
        m_out(WARNING)<<"ADCElectronicsListener::deRegisteredWithHandler() deregistered with:"<<condHandler->name()<<std::endl;
        _condHandler=NULL;
    }



    lcio::long64 ADCElectronicsListener::collectionValidSince()
    {
        return _timeStampSince;
    }



    lcio::long64 ADCElectronicsListener::collectionValidTill()
    {
        return _timeStampTill;
    }



    void ADCElectronicsListener::useLastValidIfNoDataAvailable(bool useLastValid)
    {
        _useLastValidIfNoDataAvailable=useLastValid;
    }



    bool ADCElectronicsListener::noConditionsDataAvailable()
    {
        return _noDataAvailable;
    }

}
