#include "ADCPulseConverterProcessor.h"

//C++
#include <iostream>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// STUFF needed for GEAR
#include <marlin/Global.h>
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>

using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc{

  ADCPulseConverterProcessor aADCPulseConverterProcessor;
  
  ADCPulseConverterProcessor::ADCPulseConverterProcessor() 
    : Processor("ADCPulseConverterProcessor")
  {
    // modify processor description
    _description = "ADCPulseConverterProcessor:  Convert zero-suppressed ADC raw data to pulses" ;
    
    
    // register steering parameters: name, description, class-variable, default value
    
    registerInputCollection(    LCIO::TRACKERDATA,
				"InputCollectionName" , 
				"Name of the input TrackerData collection"  ,
				_inputColName ,
				std::string("TPCData")) ;
    
//    registerOutputCollection(   LCIO::TRACKERDATA,
//				"OutputTrackerDataCollectionName" , 
//				std::string("Name of the output TrackerData collection")+
//				std::string("only the data related to a pulse will be stored"),
//				_outputDataColName ,
//				std::string("TPCDataInPulses")) ;
//
    registerOutputCollection(   LCIO::TRACKERPULSE,
				"OutputTrackerPulseCollectionName" , 
				"Name of the output TrackerPulse collection"  ,
				_outputPulseColName ,
				std::string("TPCPulses"));
    
    registerProcessorParameter( "SetOutputTransient" , 
				"if not 0 the output collection is set transient"  ,
				_outputIsTransient,
				0);

    registerProcessorParameter( "ForceSpectrumSave" , 
				"if not 0 the ADC spectrum will be stored for all pulses",
				_forceSpectrumSave,
				1);
  }

  void ADCPulseConverterProcessor::init()
  { 
    // usually a good idea to
    printParameters() ;
    
    //_ReadoutFrequency = Global::GEAR->getTPCParameters().getReadoutFrequency();
  }

  void ADCPulseConverterProcessor::processRunHeader( LCRunHeader* run)
  { 
    run->parameters().setValue( _processorName + "_revision", "$Rev:$");
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
    {
      if ( ! i->second->isOptional() || i->second->valueSet() ) 
      {
	run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
      }
    }
  } 

  void ADCPulseConverterProcessor::processEvent( LCEvent * evt )
  { 
    LCCollection* InputCollection = evt->getCollection( _inputColName );
    
    // create the output collections for pulses and zero suppressed data (ADC spectra)
    IMPL::LCCollectionVec* outputColPulse = new IMPL::LCCollectionVec(LCIO::TRACKERPULSE);
    //    IMPL::LCCollectionVec* outputColData  = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
    outputColPulse->setFlag( trkFlag.getFlag() );

    // set the transient flag as definded by the processor parameter
    outputColPulse->setTransient( (_outputIsTransient != 0) );
    //    outputColData->setTransient( (_outputIsTransient != 0) );
	
    // loop the input collection
    int nData = InputCollection->getNumberOfElements();
    for(int iData=0; iData< nData ; iData++)
    {
      TrackerData* thisData = 
		    dynamic_cast<TrackerData*>( InputCollection->getElementAt( iData ) ) ;
	    
      if (thisData)// cast is OK
      {
	TrackerPulseImpl *pulse = new TrackerPulseImpl;

	pulse->setCellID0( thisData->getCellID0() );
	pulse->setCellID1( thisData->getCellID1() );

	// now calculate time and chare from the tracker data
	pulse->setTime(    calcPulseTime( thisData) );
	pulse->setCharge(  calcPulseCharge(thisData) );

	// only save the spectrum if requested
	if (_forceSpectrumSave)
	  pulse->setTrackerData(thisData);

	outputColPulse->addElement( pulse );
      } // if thisData

    }// for iPulse

    evt->addCollection(outputColPulse, _outputPulseColName);
    //evt->addCollection(outputColData, _outputDataColName);
  }



    void ADCPulseConverterProcessor::check( LCEvent * evt ) { 


    }


    void ADCPulseConverterProcessor::end(){ 
    
    }


  float ADCPulseConverterProcessor::calcPulseTime(TrackerData* thisPulse)
  {
    // we need cellID1 to get the module, because we need the frequency, which is stored per module
    const gear::TPCModule *module;
    try{
      module = &(Global::GEAR->getTPCParameters().getModule(thisPulse->getCellID1()));
    }
    catch (gear::Exception &e)
    {
      cout << e.what() ;

      // did not find the correct module id, see if there is only one module...
      module = dynamic_cast<const gear::TPCModule*>(&(Global::GEAR->getTPCParameters().getPadLayout()));      
    }
    
    
    float timeBin=1e+9/module->getReadoutFrequency();
    
    float charge=0;
    FloatVec chargeVec=thisPulse->getChargeValues();
    for (unsigned int j=0; j<chargeVec.size(); j++)
    {
      charge+=chargeVec[j];
    }
    
    // Calculate the the time of the hit (in ns).
	double time=0;
	for(unsigned int j=0;j<chargeVec.size();j++)
	{
	  time+=chargeVec[j]*j*timeBin;
	}
	time=time/charge;
	time+=thisPulse->getTime()*timeBin;
	return time;
  }
  
  float ADCPulseConverterProcessor::calcPulseCharge(TrackerData* thisPulse)
  {
    float    PulseCharge  = 0.0;
    FloatVec ChargeValues = thisPulse->getChargeValues();
    for (FloatVec::iterator iSlice = ChargeValues.begin()++; 
	 iSlice != ChargeValues.end(); iSlice++)
    {
	    PulseCharge+=(*iSlice);
    }
    
    return PulseCharge;
  }

}
