#ifdef USE_TRACK_FITTER_KALMAN
#include "AttKalTrack.h"

#include <fstream>
#include "TPolyMarker3D.h"
#include "TCollection.h"
#include "TVector3.h"

#include "IMPL/TrackImpl.h"
#include <EVENT/LCCollection.h>
#include <streamlog/streamlog.h>
 
#include "kaldet/EXTPCKalDetector.h"
#include "kaldet/EXTPCMeasLayer.h"


using namespace IMPL;

namespace marlintpc {

AttKalTrack::AttKalTrack(const TPCParameters* tp)
:_tpcParameters(tp)
{
}

AttKalTrack::AttKalTrack(const AttKalTrack& orig)
           : _domain(orig._domain),
	     _tpcHitMap(orig._tpcHitMap),
	     _tpcParameters(orig._tpcParameters)
{
  list<ModuleRowHit>::const_iterator iter;
  for(iter=orig._hits.begin();iter!=orig._hits.end();iter++)_hits.push_back(*iter);
}

AttKalTrack::~AttKalTrack()
{
  _hits.clear();
}

void AttKalTrack::removeHits()
{
  _hits.clear();
}


TrackImpl* AttKalTrack::findTrack(ModuleRowHit firstHit)
{
    streamlog_out(DEBUG3) << "======================================" << endl
                          << "In FindTrack()"                         << endl
                          << "======================================" << endl;
    
    if(searchTrack(firstHit))
    {
       streamlog_out(MESSAGE4) << endl << "A track is found, containing "  
                               << getNumberOfHits()  << " hits" << endl;

       //save track
       TrackImpl* aTrack = new TrackImpl();
       for(list<ModuleRowHit>::iterator iter=_hits.begin();iter!=_hits.end();iter++)
       {
         TrackerHit* aHit= _tpcHitMap->getTrackerHit(*iter);
         aTrack->addHit(aHit);
       }

       //romove hits belonging to the track from map
       _tpcHitMap->removeHits(&_hits);
       return aTrack;
    }
    else
    {
       //if not find a track, remove this hit
       _tpcHitMap->removeHit(firstHit);
       return 0;
    }
}


void AttKalTrack::searchHit()//recursive
{
    streamlog_out(DEBUG3) << "---------------------------" << endl
                          << "In SearchHit()"              << endl
                          << "---------------------------" << endl;
    bool foundHit=false;

    list<ModuleRowHit> hitList= _tpcHitMap->getHitsInDomain(*_domain);
    list<ModuleRowHit>::iterator bestHit; 
    list<ModuleRowHit>::iterator iter;

    double dChi2Save = 1.e20;

    //find a best hit which has minimum delta Chi2
    for(iter=hitList.begin();iter!=hitList.end();iter++)
    {
       AttKalTrack* modTrack = create(this);//new
       if( modTrack->doAddAndFilter(*iter))
       {
          if(foundHit==false)
          { 
            bestHit=iter; 
	    dChi2Save = modTrack->getDeltaChi2();
            foundHit=true; 
          }
          else
	  {
            if( modTrack->getDeltaChi2() < dChi2Save ) {
	      bestHit=iter;
	      dChi2Save = modTrack->getDeltaChi2();
	    }
	  }
          
          TObjArray *objArrPtr = dynamic_cast<TObjArray *>(modTrack);
          if (objArrPtr) delete objArrPtr->Last();
       }
       delete modTrack;//delete
    }

    if(getType()==0 && dChi2Save>10)foundHit=false;//for Straight track

    if(foundHit)
    {
       streamlog_out(DEBUG3) << "A hit is selected: " << endl;
       doAddAndFilter(*bestHit);
       _hits.push_back(*bestHit);
       _lastBestHit = *bestHit;
       _domain->resetSkip();

       //debug:
       //printStateVector();

       streamlog_out(DEBUG3) << "Go the next hit" << endl;
       if(!_domain->setDomain(*bestHit,1))return ; 
       searchHit();
    }
    else
    {
      if(!_domain->skipDomain(_lastBestHit))
        return ;
      else
        searchHit();
    }
}


void AttKalTrack::drawEvent(LCEvent* evt, string hitInputName, int color)
{
    LCCollection *inputHits=0;
    try
    {
       inputHits=evt->getCollection(hitInputName);
    }
    catch  (DataNotAvailableException &e) {
       return;
    }  

    int hitNum=inputHits->getNumberOfElements();

    std::ofstream display("display.C",std::ios::app);
    TPolyMarker3D *polyMarkerPtr = new TPolyMarker3D( hitNum );

    polyMarkerPtr->SetBit(kCanDelete);
    polyMarkerPtr->SetMarkerColor(color);
    polyMarkerPtr->SetMarkerStyle(6);//star
       
    for(int iHit=0;iHit<hitNum;iHit++)
    {
       TrackerHit* inputHit = dynamic_cast<TrackerHit *>( inputHits->getElementAt(iHit));
       const double* pos=inputHit->getPosition();
       polyMarkerPtr->SetPoint(iHit, pos[0], pos[1], pos[2]);
    }

    polyMarkerPtr->SavePrimitive(display);
}

void AttKalTrack::drawTrack(int color)
{
   std::ofstream display("display.C",std::ios::app);
   TPolyMarker3D *polyMarkerPtr = new TPolyMarker3D( _hits.size() );
   polyMarkerPtr->SetBit(kCanDelete);
   polyMarkerPtr->SetMarkerColor(color);
   polyMarkerPtr->SetMarkerStyle(6);

   int nHits=0;
   list<ModuleRowHit>::iterator it;

   for(it=_hits.begin();it!=_hits.end(); it++)
   {
      const double* pos=TPCHitMap::getHitPosition(*it);
      polyMarkerPtr->SetPoint(nHits++, pos[0], pos[1], pos[2]);
   }

   polyMarkerPtr->SavePrimitive(display);
}

}
#endif
