#include "RowBasedHitFinderProcessor.h"

//C++
#include <iostream>
#include <algorithm>

//LCIO
#include "EVENT/LCFloatVec.h"
#include "IMPL/TrackerDataImpl.h"
#include "IMPL/TrackerPulseImpl.h"
#include "IMPL/TrackerHitImpl.h"
#include "IMPL/LCFlagImpl.h"

// GEAR
#include "gear/TPCParameters.h"
#include "gear/TPCModule.h"
#include "gearimpl/GlobalPadIndex.h"

//MARLIN
#include "marlin/Global.h"

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"
// the common flag word definitions in MarlinTPC
//#include "FlagwordDefinitions.h"

using namespace std;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc
{

RowBasedHitFinderProcessor aRowBasedHitFinderProcessor;


RowBasedHitFinderProcessor::RowBasedHitFinderProcessor() : Processor("RowBasedHitFinderProcessor")
{
    _description = "RowBasedHitFinderProcessor calculates TrackerHits from TrackerPulses" ;

    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(LCIO::TRACKERPULSE,
                            "InputTrackerPulses" ,
                            "Name of the input TrackerPulses collection"  ,
                            _inputTrackerPulsesCollectionName ,
                            string("TPCPulses")) ;

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputTrackerHits" ,
                             "Name of the output TrackerHits collection"  ,
                             _outputTrackerHitsCollectionName ,
                             string("TPCHits")) ;

    registerOptionalParameter("InputTPCConditions",
                              "Optional: Name of the input collection containing the TPC conditions data",
                              _inputTPCConditionsCollectionName,
                              TPCConditions::getDefaultColName()) ;

    registerProcessorParameter("WriteOutputToStorage" ,
                               "if true, the output hits collection is written to storage (default: true)"  ,
                               _outputHitsPersistent,
                               true);

    registerProcessorParameter("MaxEmptyPadsOverride" ,
                               "Maximum number of empty consecuitve pads in hit (default: 1)"  ,
                               _maxEmptyPadsOverride ,
                               1) ;

    registerProcessorParameter("MinHitSizeOverride" ,
                               "Minimum size of hit / Minimum number of Pads (default: 1)"  ,
                               _minHitSizeOverride ,
                               1) ;

    registerProcessorParameter("MinMaxPulseSliceHeightOverride",
                               "Minimum size of the maximum slice value of the maximal pulse (in charge) in a hit.",
                               _minMaxPulseSliceHeightOverride,
                               float(12.)) ;

    registerProcessorParameter("MaxTimeSpreadOverride" ,
                               "Maximum time between pulses in a hit in ns (default: 200.)"  ,
                               _maxTimeSpreadOverride ,
                               float(200.)) ;

    registerOptionalParameter("VDriftOverride" ,
                              "Optional: Set drift velocity in case there is no conditions data in mm/mu s"  ,
                              _driftVelocityOverride ,
                              float(0.)) ;

    registerOptionalParameter("DeadPadsOverride",
                              "Optional: the gear pad indices of dead channels",
                              _deadPadIndicesOverride,
                              _deadPadIndicesOverride) ;

    registerOptionalParameter("ChargeConversionFactorOverride" ,
                              "Optional: The charge conversion factor from ADC values to primary electrons",
                              _chargeConversionFactorOverride ,
                              float(1.)) ;

    registerOptionalParameter("UseGlobalCoordinates",
                              "Optional: If set to true, the output coordinate system is changed to be the global coordinates (default is false).",
                              _useGlobalCoord ,
                              false);

}



void RowBasedHitFinderProcessor::init()
{
    printParameters();

    // get drift velicity, either measured value from calibration
    // or nominal value given as process parameter
    if (parameterSet("VDrift"))
        {
            m_out(WARNING) << "the drift velocity is taken from the steering file. please change" << endl;
        }
}



void RowBasedHitFinderProcessor::processRunHeader(LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 1842 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (!i->second->isOptional() || i->second->valueSet())
                run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
        }
}



void RowBasedHitFinderProcessor::processEvent(LCEvent* evt)
{
    /** the processor is supposed to find "hits" in "rows"
    input are pulses: (charge,time,moduleID,cellID + qualityFlag)
    output are hits: ( [x,y,z] + covariance matrix, charge + qualityFlag )

    steering input: - minimal hit size
                    - max number of dead/empty pads within a hit
            - max distance in z/time between pulses
    external input: - dead pads
                    - drift velocity
            - charge conversion factor


        structurally this happens in a decomposition of:
        (A) sort the pulses by module & row
        (B) LOOP: look for pulses that belong together in a row (for all rows in a module, for all modules)
        (C) create the "Hit" by calculating all numbers

    **/

    // get the input pulse collection
    LCCollection* inputPulses;
    try
        {
            inputPulses = evt->getCollection(_inputTrackerPulsesCollectionName);
        }
    catch (DataNotAvailableException &e)
        {
            // if no pulses are available, skip to next event
            m_out(DEBUG) << "No pulses available in event " << evt->getEventNumber() << endl;
            return;
        }

    // create output hit collection and set the persistency flag
    LCCollectionVec* tpcHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
    tpcHitCollection->setTransient(!_outputHitsPersistent);

    // the main parameter for the geometry description is the gear parameter:
    const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();

    /** ********************************************************************************
    the basic data structure for the hit search is a single row;
    the row is a simple vector of the Pulse pointers:
    vector<TrackerPulse*>

    this is repeated for every (unique) row on a module -> use a map of these
    repeat this for every possible module, since the module id is unique use a map

    so the final data structure for the hit search is:
    map<int moduleID, map<int rowID, vector<TrackerPulse*>* >
    (no typedef is introduced, although arguably this might improve the readability)

    a helper class is instantiated to hold the hit candidates: "hitCandidate"
    the temporary result of the hit candidate finding is stored in a vector of these
    the key is again a map -- since it is stored per /unique/ module


    ********************************************************************************
     **/

    map<int, map<int, vector<TrackerPulse*>*>*> theRowSortedPulsesPerModule;
    map<int, vector<hitCandidate> >             theHitCandidatesPerModule;

    // (A): sort pulses
    _sortPulses(inputPulses, tpcParameters, theRowSortedPulsesPerModule);

    // (B)
    _findHitCandidates(theRowSortedPulsesPerModule, tpcParameters, theHitCandidatesPerModule);

    // (C)
    _createHitCollection(theHitCandidatesPerModule, tpcParameters, tpcHitCollection);

    evt->addCollection(tpcHitCollection, _outputTrackerHitsCollectionName);

    m_out(MESSAGE) << "Found " << tpcHitCollection->getNumberOfElements() << " hits in the event." << endl;

    m_out(DEBUG) << "There are " << inputPulses->getNumberOfElements()
    << " pulses to begin with, and a total of "
    << tpcHitCollection->getNumberOfElements() << " hits found." << endl;
}



void RowBasedHitFinderProcessor::check(LCEvent* evt)
{
}



void RowBasedHitFinderProcessor::end()
{
}



void RowBasedHitFinderProcessor::_sortPulses(LCCollection* inputPulses, const gear::TPCParameters& tpcParameters,
                                             map<int, map<int, vector<TrackerPulse*>* >* >& rowSortedPulsesPerModule)
{
    /**  put the pulses in a real container
     in principle the idea is easy:
       every unique row on a unique module gets another pulse added
       keep in mind that more than one pulse can occur on a single pad!

     practically this tends to be a bit /lengthy/ to write in stl containers,
     please refer to the stl documentation
    **/

    pair<map<int, map<int, vector<TrackerPulse*>*>*>::iterator, bool> moduleInserter;
    pair<map<int, vector<TrackerPulse*>*>::iterator, bool> rowInserter;

    for (int aPulse = 0; aPulse < inputPulses->getNumberOfElements(); ++aPulse)
        {
            TrackerPulse* inputPulse = dynamic_cast<TrackerPulse*>(inputPulses->getElementAt(aPulse));
            if (inputPulse) // the cast worked
                {
                    const int moduleID         = inputPulse->getCellID1();
                    const int padID            = inputPulse->getCellID0();

                    const TPCModule& theModule = tpcParameters.getModule(moduleID);
                    const int rowID            = theModule.getRowNumber(padID);

                    // first create/use the module insertion iterator
                    moduleInserter =
                        rowSortedPulsesPerModule.insert(pair<int,  map<int, vector<TrackerPulse*>* >* >
                                                        (moduleID, new map<int, vector<TrackerPulse*>* >));
                    // use the newly created/already existing map pointer to insert the row
                    rowInserter = ((moduleInserter.first)->second)->insert
                                  (pair<int, vector<TrackerPulse*>* >(rowID, new vector<TrackerPulse*>));

                    // and finally add the Pulse itself
                    ((rowInserter.first)->second)->push_back(inputPulse);
                }
        } // end input pulse loop
}



void RowBasedHitFinderProcessor::_findHitCandidates
(map<int, map<int, vector<TrackerPulse*>* >* >& rowPulsesPerModule, const gear::TPCParameters& tpcParameters,
 map<int, vector<hitCandidate> >& theHitCandidateMap)
{
    /**  now search in each row (of each module)
     a) for the highest pulse
     b) look for neighboring pulses
     c) store it in a candidate
     d) remove the stored pulses and redo from a)
    **/

    // two iterator types are needed: for the modules and for the rows
    map<int, map<int, vector<TrackerPulse*>* >* >::iterator aModuleWithPulses;
    map<int, vector<TrackerPulse*>* >::iterator aRowWithPulses;
    vector<TrackerPulse*>::iterator pulseRow;

    // outer loop:: the modules
    for (aModuleWithPulses = rowPulsesPerModule.begin();
            aModuleWithPulses != rowPulsesPerModule.end(); ++aModuleWithPulses)
        {
            int moduleID;
            vector<hitCandidate> theHitCandidates;

            // inner loop:: the rows
            for (aRowWithPulses  = aModuleWithPulses->second->begin();
                    aRowWithPulses != aModuleWithPulses->second->end(); ++aRowWithPulses)
                {
                    m_out(DEBUG) << "Starting search of neighbouring pulses in in row: "
                    << aRowWithPulses->first << endl;

                    // now row wise, decrease the pulses consecutively
                    while (aRowWithPulses->second->size() != 0)
                        {
                            // first find the pulse with the highest charge -- the seed
                            TrackerPulse* theSeedPulse = _findHighestPulseInRow(*(aRowWithPulses->second));
                            moduleID = theSeedPulse->getCellID1();

                            m_out(DEBUG) << "Found a highest pulse with charge = " << theSeedPulse->getCharge()
                            << " on pad " << theSeedPulse->getCellID0()
                            << " and with time " << theSeedPulse->getTime() << endl;
                            // time saver, conclude the search,
                            // if the maximal pulse height is below the minimum requirement
                            if (!_maxPulseHeightCheck(theSeedPulse))
                                {
                                    m_out(DEBUG) << "Pulse is rejected, since it is too small. Continue in next row. " << endl;
                                    break;
                                }
                            hitCandidate theProtoHit(theSeedPulse);
                            bool borderFlagIsSet = false;

                            vector<TrackerPulse*> theNeighbourPulses =
                                _getNeighbourPulses(theSeedPulse, tpcParameters, *(aRowWithPulses->second), borderFlagIsSet);

                            if (borderFlagIsSet)
                                theProtoHit.setBoundaryFlag();

                            for (vector<TrackerPulse*>::iterator hitPulse = theNeighbourPulses.begin();
                                    hitPulse < theNeighbourPulses.end(); ++hitPulse)
                                {
                                    theProtoHit.addPulse(*hitPulse);
                                }
                            theHitCandidates.push_back(theProtoHit);
                        }
                }
            theHitCandidateMap.insert(pair<int, vector<hitCandidate> >(moduleID, theHitCandidates));

            m_out(DEBUG) << "On module " << aModuleWithPulses->first
            << " there are " << theHitCandidates.size() << " hits found." << endl;

        } // end outer loop  of modules
    return;
}



TrackerPulse* RowBasedHitFinderProcessor::_findHighestPulseInRow(vector<TrackerPulse*>& pulsesInARow)
{
    TrackerPulse* highestPulse = pulsesInARow.at(0);

    vector<TrackerPulse*>::iterator eraser = pulsesInARow.begin();
    for (vector<TrackerPulse*>::iterator iter = pulsesInARow.begin(); iter < pulsesInARow.end(); ++iter)
        {
            if (fabs((*iter)->getCharge()) > fabs(highestPulse->getCharge()))
                {
                    highestPulse = *iter;
                    eraser = iter;
                }
        }
    pulsesInARow.erase(eraser);
    return highestPulse;
}



bool RowBasedHitFinderProcessor::_maxPulseHeightCheck(TrackerPulse* theSeedPulse)
{
    EVENT::TrackerData*  trackerData = theSeedPulse->getTrackerData();
    vector<float> theADCSpectrum = trackerData->getChargeValues();

    return *max_element(theADCSpectrum.begin(), theADCSpectrum.end()) >= _getMinMaxPulseSliceHeight();
}



vector<TrackerPulse*> RowBasedHitFinderProcessor::_getNeighbourPulses
(TrackerPulse* theSeedPulse,  const gear::TPCParameters& tpcParameters,
 vector<TrackerPulse*>& pulsesInARow, bool& borderflag)
{
    vector<TrackerPulse*> theReturnPulses;
    theReturnPulses.clear();

    const TPCModule& theModule = tpcParameters.getModule(theSeedPulse->getCellID1());
    int   thePadIndex  = theSeedPulse->getCellID0();
    float theTime      = theSeedPulse->getTime();
    int numberOfMisses = 0;

    bool finishedLeft = false; // arbitrary choice: start with searching "left", used to change search direction
    bool stopSearch   = false; // loop control

    // start searching for neighbour pulses
    while (!stopSearch)
        {
            try
                { // to get the index of the left next pad
                    if (!finishedLeft)
                        thePadIndex = theModule.getLeftNeighbour(thePadIndex);
                    else
                        thePadIndex = theModule.getRightNeighbour(thePadIndex);
                }
            catch (...) // FIXME: what kind of exception is thrown?
                {
                    m_out(DEBUG) << "The hit candidate is at the border!" << endl;
                    borderflag = true;

                    if (!finishedLeft)
                        {
                            finishedLeft = true;
                            thePadIndex  = theSeedPulse->getCellID0();
                        }
                    else
                        stopSearch = true;
                }

            // if more than one pulse matches the requirements, the best match needs to be found
            // store this in a temporary vector that is searched for the best match
            vector<vector<TrackerPulse*>::iterator> tempPulses;

            for (vector<TrackerPulse*>::iterator iter = pulsesInARow.begin(); iter < pulsesInARow.end(); ++iter)
                {
                    if (((*iter)->getCellID0() == thePadIndex) && (fabs((*iter)->getTime() - theTime) <  _getMaxTimeSpread()))
                        tempPulses.push_back(iter);
                }

            if (tempPulses.size() ==  0)
                numberOfMisses++;
            else
                {
                    vector<TrackerPulse*>::iterator bestPulseMatch = _findBestNeighbourPulseMatch(tempPulses, theTime);

                    theReturnPulses.push_back(*bestPulseMatch);
                    theTime = (*bestPulseMatch)->getTime();
                    pulsesInARow.erase(bestPulseMatch);
                    tempPulses.clear();
                }

            if (numberOfMisses > _getMaxEmptyPads())
                {
                    if (!finishedLeft)
                        {
                            finishedLeft   = true;
                            theTime        = theSeedPulse->getTime();
                            numberOfMisses = 0;
                            thePadIndex    = theSeedPulse->getCellID0();
                        }
                    else
                        stopSearch = true;
                }
        }
    m_out(DEBUG) << "A hit candidate with a width of " << theReturnPulses.size() << " is found." << endl;
    return theReturnPulses;
}



vector<TrackerPulse*>::iterator RowBasedHitFinderProcessor::_findBestNeighbourPulseMatch
(vector<vector<TrackerPulse*>::iterator>& pulseMatches, float theTime)
{
    if (pulseMatches.size() == 1)
        return pulseMatches.at(0);
    else
        {
            vector<TrackerPulse*>::iterator bestMatch = *(pulseMatches.begin());
            float minTime = fabs((*bestMatch)->getTime() - theTime);

            for (vector<vector<TrackerPulse*>::iterator>::iterator searchIter = pulseMatches.begin();
                    searchIter < pulseMatches.end(); ++searchIter)
                {
                    if (fabs((**searchIter)->getTime() - theTime) < minTime)
                        {
                            minTime = fabs((**searchIter)->getTime() - theTime);
                            bestMatch = *searchIter;
                        }
                }
            m_out(DEBUG) << "In the hit more than one pulse was found to match to the protohit,"
            "the best match in time was chosen." << endl;
            return bestMatch;
        }
}



void RowBasedHitFinderProcessor::_createHitCollection
(map<int, vector<hitCandidate> >& theHitCandidateMap, const gear::TPCParameters& tpcParameters,
 LCCollectionVec* tpcHitCollection)
{
    /** now loop over all hits on all modules and:
    1) check their validity
    2) calculate the geometry dependent position of the fixed coordinate given by the row (y or r)
    3) calculate the geometry dependent position of the free coordinate (x or phi)
    4) calculate the z coordinate
    5) calculate the error matrix, which is diagonal in the local coordinate system
    6) store the TrackerHit in the output collection
    **/

    map<int, vector<hitCandidate> >::iterator hitCandidatesPerModule;

    // loop over modules
    for (hitCandidatesPerModule = theHitCandidateMap.begin();
            hitCandidatesPerModule != theHitCandidateMap.end(); ++hitCandidatesPerModule)
        {
            const gear::TPCModule &module =
                marlin::Global::GEAR->getTPCParameters().getModule((*hitCandidatesPerModule).first);
            //marlin::Global::GEAR->getTPCParameters().getModule(currentPulse->first.getModuleID () );

            //loop over hitCandidates:
            for (vector<hitCandidate>::iterator candidate = hitCandidatesPerModule->second.begin();
                    candidate < hitCandidatesPerModule->second.end(); ++candidate)
                {
                    // check validity
                    if (_hitIsValid(*candidate) == false)
                        continue;

                    // change output coordinate system
                    double position[3];
                    // make the error calculation easier:
                    double meanFreeCoordinate;
                    if (_useGlobalCoord)
                        {
                            gear::Vector2D positionGlobal2D = module.localToGlobal(
                                                                  _calculateFixedCoordinate(*candidate, tpcParameters),
                                                                  _calculateFreeCoordinate(*candidate, tpcParameters));
                            position[0] = positionGlobal2D[0];
                            position[1] = positionGlobal2D[1];
                            position[2] = _calculateZCoordinate(*candidate);
                            meanFreeCoordinate = position[1];
                        }
                    else
                        {
                            meanFreeCoordinate = _calculateFreeCoordinate(*candidate, tpcParameters);
                            position[0] = meanFreeCoordinate;
                            position[1] = _calculateFixedCoordinate(*candidate, tpcParameters);
                            position[2] = _calculateZCoordinate(*candidate);
                        }

                    float covMatrix[6] = {pow(_calculateFreeError(*candidate, meanFreeCoordinate, tpcParameters), 2) , 0., 0.,
                                          pow(_calculateFixedError(*candidate, tpcParameters), 2) , 0.,
                                          pow(_calculateZError(*candidate), 2)
                                         };

                    TrackerHitImpl* theHit = new TrackerHitImpl();
                    //      theHit->setType(int type) ???
                    theHit->setPosition(position);
                    theHit->setCovMatrix(covMatrix);
                    theHit->setEDep(_convertCharge(*candidate));
                    theHit->setEDepError(sqrt(_convertCharge(*candidate)));

                    m_out(DEBUG) << "The hit coordinates are [" << position[0] << "|" << position[1] << "|" << position[2] << "]" << endl
                    << " with variances [" << theHit->getCovMatrix().at(0) << "|" << theHit->getCovMatrix().at(3) << "|" <<  theHit->getCovMatrix().at(5)  << "]" << endl
                    << "and the charge is " << theHit->getEDep() << " +- " << theHit->getEDepError() << endl;

                    vector<TrackerPulse*> hitPulses = candidate->getPulses();
                    for (vector<TrackerPulse*>::const_iterator aHitPulse = hitPulses.begin();
                            aHitPulse < hitPulses.end(); ++aHitPulse)
                        {
                            theHit->rawHits().push_back(*aHitPulse);
                        }

                    // the time is inherited from the pulse with the largest charge
                    theHit->setTime(candidate->getMaximumPulse()->getTime());

                    tpcHitCollection->addElement(theHit);
                }
        }

    return;
}



bool RowBasedHitFinderProcessor::_hitIsValid(const hitCandidate& cand)
{
    return (int)cand.getWidth() >= _getMinHitSize();
}



double RowBasedHitFinderProcessor::_calculateFixedCoordinate(const hitCandidate& cand, const gear::TPCParameters& tpcParameters)
{
    const TrackerPulse* maxPulse = cand.getMaximumPulse();

    const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();
    Vector2D padCenterLocal = localPadPlane.getPadCenter(maxPulse->getCellID0());

    if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
        return padCenterLocal[1]; // this is truly y
    else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
        return padCenterLocal[0]; // why has this to be a different index in polar... -> this is r, anyway
    else
        throw gear::Exception("Coordinate type unknown!");
}



double RowBasedHitFinderProcessor::_calculateFreeCoordinate(const hitCandidate& cand, const gear::TPCParameters& tpcParameters)
{
    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

    vector<TrackerPulse*> thePulses = cand.getPulses();

    double chargeSum             = 0.;
    double weightedSum           = 0.;
    double position              = 0.;
    double weightedCorrectionSum = 0.;
    const double localPI         = 3.41416; // i know, this is just a local Pi version, I don't know which math lib is available
    Vector2D padCenterLocal;
    for (vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
        {
            chargeSum += (*aPulse)->getCharge();
            padCenterLocal = localPadPlane.getPadCenter((*aPulse)->getCellID0());

            if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
                position = padCenterLocal[0]; // cartesian x
            else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
                {
                    position = padCenterLocal[1]; // polar phi
                    // possible trouble: phi definition range is [-Pi,+Pi]
                    // for the sake of computation: first shift to [Pi,3*Pi] and then back again
                    position += 3.*localPI;
                }
            else
                throw gear::Exception("Coordinate type unknown!");

            weightedSum += (*aPulse)->getCharge() * position;
            weightedCorrectionSum += (*aPulse)->getCharge() * 3.*localPI;
        }

    // now correct back for polar coordinates:
    if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
        weightedSum -= weightedCorrectionSum;

    if (chargeSum > 0.)
        return weightedSum / chargeSum;
    else
        return 0.;
}



double RowBasedHitFinderProcessor::_calculateZCoordinate(const hitCandidate& cand)
{
    return _getDriftVelocity() *(cand.getMaximumPulse())->getTime() / 1000.; // time is in [ns], drift velocity in [mm/mus]
}



double RowBasedHitFinderProcessor::_calculateFreeError(const hitCandidate& cand, double freeMean, const gear::TPCParameters& tpcParameters)
{
    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

    // returns automatically the right coordinate system
    // The width of the pad at padIndex in mm (CARTESIAN) or radians (POLAR).
    double padWidth = module.getPadWidth(maxPulse->getCellID0());

    const double SQRT12 = 3.46410161513775458705;
    double position     = 0.;
    double freeError    = 0.;

    vector<TrackerPulse*> thePulses = cand.getPulses();
    Vector2D padCenterLocal;

    if (cand.getWidth() == 1)
        return padWidth / SQRT12; // it's the pad width divided by sqrt(12)
    else
        {
            for (vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
                {
                    padCenterLocal = localPadPlane.getPadCenter((*aPulse)->getCellID0());

                    if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
                        position = padCenterLocal[0]; // cartesian x
                    else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
                        {
                            position = padCenterLocal[1]; // polar phi
                        }
                    else
                        throw gear::Exception("Coordinate type unknown!");
                    freeError += pow((position - freeMean) , 2);
                }
        }
    const double theTotalCharge = cand.getCharge();

    return sqrt(freeError) / theTotalCharge;
}



double RowBasedHitFinderProcessor::_calculateFixedError(const hitCandidate& cand, const gear::TPCParameters& tpcParameters)
{
    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    const gear::TPCModule& module = tpcParameters.getModule(maxPulse->getCellID1());

    const double SQRT12 = 3.46410161513775458705;

    int rowNumber    = module.getRowNumber(maxPulse->getCellID0());
    double rowHeight = module.getRowHeight(rowNumber);

    return rowHeight / SQRT12;
}



double RowBasedHitFinderProcessor::_calculateZError(const hitCandidate& cand)
{
    // for now the z position is calculated simply from the time of the maximum pulse
    // therefore the error on z is the same as for the pulseMatches

    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    double pulseTimeError = sqrt((maxPulse->getCovMatrix()).at(2));

    return _getDriftVelocity() * pulseTimeError / 1000.;
}



double RowBasedHitFinderProcessor::_convertCharge(const hitCandidate& cand)
{
    return _getChargeConversionFactor() *(cand.getCharge());
}



RowBasedHitFinderProcessor::hitCandidate::hitCandidate(TrackerPulse* aPulse)
        :  _maxPulse(NULL), _maxPulseCharge(0.0), _numberOfDeadPads(0), _isAtBoundary(false)
{
    _thePulses.clear();
    addPulse(aPulse);
}



void RowBasedHitFinderProcessor::hitCandidate::addPulse(TrackerPulse* aPulse)
{
    _thePulses.push_back(aPulse);
    if (fabs(aPulse->getCharge()) > fabs(_maxPulseCharge))
        {
            _maxPulseCharge = aPulse->getCharge();
            _maxPulse       = aPulse;
        }
}


void RowBasedHitFinderProcessor::hitCandidate::addDeadPad()
{
    _numberOfDeadPads++;
}



double RowBasedHitFinderProcessor::hitCandidate::getCharge() const
{
    double tmpCharge = 0.0;
    for (vector<TrackerPulse*>::const_iterator aPulse = _thePulses.begin(); aPulse < _thePulses.end(); ++aPulse)
        tmpCharge += (*aPulse)->getCharge();
    return tmpCharge;
}



double RowBasedHitFinderProcessor::hitCandidate::getMaxPulseHeight()
{
    return _maxPulseCharge;
}



unsigned int RowBasedHitFinderProcessor::hitCandidate::getWidth() const
{
    return _thePulses.size();
}



void RowBasedHitFinderProcessor::hitCandidate::setBoundaryFlag()
{
    _isAtBoundary = true;
}



bool RowBasedHitFinderProcessor::_pulseComparator(TrackerPulse* p0, TrackerPulse* p1)
{
    return fabs(p0->getCharge()) < fabs(p1->getCharge());
}



int RowBasedHitFinderProcessor::_getMaxEmptyPads()
{
    if (parameterSet("MaxEmptyPadsOverride"))
        {
            return _maxEmptyPadsOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMaxEmptyPads()  Exception: number can't be found!"));
        }
}



int RowBasedHitFinderProcessor::_getMinHitSize()
{
    if (parameterSet("MinHitSizeOverride"))
        {
            return _minHitSizeOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMinHitSize()  Exception: Number can't be found!"));
        }
}



double RowBasedHitFinderProcessor::_getMinMaxPulseSliceHeight()
{
    if (parameterSet("MinMaxPulseSliceHeightOverride"))
        {
            return _minMaxPulseSliceHeightOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMinMaxPulseSliceHeight()  Exception: Number can't be found!"));
            return 0.;
        }
}



double RowBasedHitFinderProcessor::_getMaxTimeSpread()
{
    if (parameterSet("MaxTimeSpreadOverride"))
        {
            return _maxTimeSpreadOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMaxTimeSpread()  Exception: Number can't be found!"));
            return 0.;
        }
}



double RowBasedHitFinderProcessor::_getDriftVelocity()
{
    if (parameterSet("VDriftOverride"))
        {
            return _driftVelocityOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getVDrift()  Exception: Number can't be found!"));
        }
}



const vector<int>& RowBasedHitFinderProcessor::_getDeadPads()
{
    if (parameterSet("DeadPadsOverride"))
        {
            return _deadPadIndicesOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getDeadPads()  Exception: Number can't be found!"));
        }
}



double RowBasedHitFinderProcessor::_getChargeConversionFactor()
{
    if (parameterSet("ChargeConversionFactorOverride"))
        {
            return _chargeConversionFactorOverride;
        }
    else
        {
            throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getChargeConversionFactor()  Exception: Number can't be found!"));
            return 0.;
        }
}
}// namespace marlintpc
