#ifndef PHOTOELECTRICPULSEPROCESSOR_H
#define PHOTOELECTRICPULSEPROCESSOR_H 1

#include <string>
#include <vector>

#include "marlin/Processor.h"

namespace EVENT {

	class LCRunHeader;
	class LCEvent;
}

namespace marlintpc {

/** PhotoelectricPulseProcessor : This processor takes a file containing positions of photosensitive dots and lines and simulates them being hit by a pulse of photons
 * 
 *    @author J. Abernathy (UVic)<br>
 */
	class PhotoelectricPulseProcessor : public marlin::Processor {

public:
	/**
	 * constructor
	 */
	PhotoelectricPulseProcessor();

	/** destructor
	 */
	virtual ~PhotoelectricPulseProcessor();

	/** used by Marlin to create a new processor
	 */
	virtual Processor*  newProcessor() { return new PhotoelectricPulseProcessor() ; }

	/** Called at the begin of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms.
	 */
	virtual void init() ;

	/** Called for every run.
	 */
	virtual void processRunHeader( EVENT::LCRunHeader* run ) ;

	/** Called for every event - the working horse.
	 */
	virtual void processEvent( EVENT::LCEvent * evt ) ;

	/** Called after the event has been processed
	 */
	virtual void check( EVENT::LCEvent * evt ) ;

	/** Called after data processing for clean up.
	 */
	virtual void end() ;

protected:

	struct Photodot {

		Photodot(double x, double y, double diameter):
			_x(x),
			_y(y),
			_diameter(diameter){};
		double _x;
		double _y;
		double _diameter;
	};

	struct Photoline {

		Photoline(double x1, double y1, double x2, double y2, double width):
			_x1(x1),
			_y1(y1),
			_x2(x2),
			_y2(y2),
			_width(width) {};
		double _x1;
		double _y1;
		double _x2;
		double _y2;
		double _width;
	};

private:

	/** the name of the file containing the positions of the photodots and photolines
	 */
	std::string _photogeometry_filename;

	/** the name of the output LCGenericObject collection from this processor containing electron clouds
	 */
	std::string _output_collection_name;

	/** the name of the output collection which will contain the SimTrackerHits
	 */
	std::string _output_simtrackerhit_collection_name;

	/** the mean number of electrons to create per photodot
	*/
	double _mean_electrons_per_photodot;

	/** the initial z position of the electron clouds
	*/
	double _initial_z_position;

	/** the collection of photodots (assume they will stay the same for the entire run)
	 */
	std::vector<Photodot> _photodots;

	/** the collection of photolines (assume they will stay the same for the entire run)
	 */
	std::vector<Photoline> _photolines;
};

} // namespace marlintpc

#endif // PHOTOELECTRICPULSEPROCESSOR_H
