#include "DrawMagneticFieldProcessor.h"
#include <iostream>
#include <string>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>
#include <Exceptions.h>

#include "GlobalFieldProcessor.h"
#include "CLHEP/Vector/ThreeVector.h"

#include "TAxis.h"
#include "TApplication.h"
#include "TCanvas.h"
#include "TColor.h"
#include "TGraph2D.h"
#include "TPaletteAxis.h"
#include "TROOT.h"
#include "TStyle.h"
#include "TSystem.h"
#include "TMath.h"

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace CLHEP;


namespace marlintpc
{

DrawMagneticFieldProcessor aDrawMagneticFieldProcessor;


DrawMagneticFieldProcessor::DrawMagneticFieldProcessor() :
		Processor("DrawMagneticFieldProcessor"),
		_draw(true)
{
  // the processor description
  _description = "DrawMagneticFieldProcessor draws plots of the Br and Bz components of the loaded field maps";
  
  /* register steering parameters: name, description, class-variable, default value  */
  registerOptionalParameter(
  		"DrawFieldMap",
  		"The name of this section of the TPC.",
  		_draw,
  		true);

  registerOptionalParameter(
    	"xMin",
    	"The lower x range of the plots in mm",
    	_xMin,
    	-600.);

  registerOptionalParameter(
      	"xMax",
      	"The upper x range of the plots in mm",
      	_xMax,
      	600.);

  registerOptionalParameter(
      	"xBins",
      	"The number of bins in x of the plots",
      	_xBins,
      	30);

  registerOptionalParameter(
    	"zMin",
    	"The lower z range of the plots in mm",
    	_zMin,
    	-800.);

  registerOptionalParameter(
      	"zMax",
      	"The upper z range of the plots in mm",
      	_zMax,
      	800.);

  registerOptionalParameter(
      	"zBins",
      	"The number of bins in z of the plots",
      	_zBins,
      	40);

}


void DrawMagneticFieldProcessor::init()
{ 
  m_out(DEBUG) << " init() called  " << endl;
  
  // you might want to list the used parameters by invoking
  printParameters();

  if (_xMin > _xMax) {
	  double temp = _xMin;
	  _xMin = _xMax;
	  _xMax = temp;
  }

  if (_zMin > _zMax) {
	  double temp = _zMin;
  	  _zMin = _zMax;
  	  _zMax = temp;
  }

}


void DrawMagneticFieldProcessor::processEvent(LCEvent* evt)
{ 
	if (!_draw) {
		return;
	}

	// need to create a ROOT application to draw a canvas, it does not need to be run
	TApplication* app = new TApplication("app", new int(0), new char* );
	app->SetReturnFromRun( kFALSE );
	if (gROOT->IsBatch()) {
		gROOT->SetBatch( kFALSE );
	}

	// fill two graphs for Br and Bz
	TGraph2D* graphBr = new TGraph2D();
	TGraph2D* graphBz = new TGraph2D();
	int counter = 0;
	for (int xBin = 0; xBin < _xBins; ++xBin) {
		for (int zBin = 0; zBin < _zBins; ++zBin) {
			// calculate the position
			Hep3Vector position = Hep3Vector( xBin*(_xMax-_xMin)/_xBins+_xMin, 0., zBin*(_zMax-_zMin)/_zBins+_zMin );

			// get the magnetic field at the position
			// need to use the GlobalFieldProcessor
			Hep3Vector field = GlobalFieldProcessor::get_instance().get_global_magnetic_field( position );

			graphBr->SetPoint( counter, position.x(), position.z(), field.rho() );
			graphBz->SetPoint( counter, position.x(), position.z(), field.z() );
			++counter;
		}
	}

	// some styling to axis and a nicer color palette
	gROOT->SetStyle( "Plain" );
	gStyle->SetPalette( 1 );
	const Int_t NRGBs = 5;
    const Int_t NCont = 99;
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.00, 0.87, 1.00, 0.51 };
    Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
    Double_t blue[NRGBs]  = { 0.51, 1.00, 0.12, 0.00, 0.00 };
    TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
    styleHist(graphBr, "B_{r} [T]", "x [mm]", "z [mm]", "B_{r} [T]");
    styleHist(graphBz, "B_{z} [T]", "x [mm]", "z [mm]", "B_{z} [T]");

    // draw a canvas
	TCanvas* canv = new TCanvas( "Magnetic Field Map", "Magnetic Field Map", 1400, 700 );
    canv->Divide( 2, 1 );
    canv->GetPad(1)->SetRightMargin(0.12);
    canv->GetPad(1)->SetLeftMargin(0.08);
    canv->GetPad(2)->SetRightMargin(0.12);
    canv->GetPad(2)->SetLeftMargin(0.08);
    canv->cd(1);
	graphBr->Draw( "cont4z" );
	canv->cd(2);
	graphBz->Draw( "cont4z" );
	gPad->Update();
	gPad->Update();

	// need some wait loop to keep canvas open
	std::cout << "Press return to continue ..." << std::endl;
	int flag = fcntl(1, F_GETFL, 0);
    int key = 0;
    while(true) {
        gSystem->ProcessEvents();
        fcntl(1, F_SETFL, flag | O_NONBLOCK);
        key = getchar();
	        if((key == '\n') || (key == '\r'))
            break;
	        usleep(1000);
    }
    fcntl(1, F_SETFL, flag);

    // draw only once
    _draw = false;
    return;
}


void DrawMagneticFieldProcessor::styleHist(TGraph2D* g, string title, string xTitle, string yTitle, string zTitle) {
	g->GetXaxis()->SetTitle( xTitle.c_str() );
    g->GetXaxis()->SetLabelSize(0.02);
    g->GetXaxis()->SetTitleSize(0.03);
    g->GetYaxis()->SetTitle( yTitle.c_str() );
    g->GetYaxis()->SetLabelSize(0.02);
    g->GetYaxis()->SetTitleSize(0.03);
    g->GetZaxis()->SetTitle( zTitle.c_str() );
    g->GetZaxis()->SetLabelSize(0.02);
    g->GetZaxis()->SetTitleSize(0.03);
    g->SetTitle( title.c_str() );
}
} // close namespace brackets

