#ifndef Pedestal_h
#define Pedestal_h 1

//C++
#include <iostream>
#include <string>
#include <utility>

//LCIO
#include "lcio.h"
#include "UTIL/LCFixedObject.h"

#define PedestalNINTVals 2       // DAQ channel number and readout module ID here called channelID and readoutGroup
#define PedestalNFLOATVals 2     // Pedestal Value and Pedestal Width
#define PedestalNDOUBLEVals 0

namespace tpcconddata
{

    class Pedestal;

    typedef Pedestal LCPedestal;

    /** Class that combines information on pedestals for storage.
     *  The channel ID and the readout module ID is compound with the
     *  Pedestal value and width into a single object of type LCGenericObject.
     *  Based on the LCFixedObject template.
     *  This can be used to be stored in a LCCollectionVec.
     * @author M. E. Janssen (DESY), R. Diener (DESY)
     */
    
    class Pedestal : public UTIL::LCFixedObject<PedestalNINTVals,PedestalNFLOATVals,PedestalNDOUBLEVals>
    {

        public:

            /** Convenient constructor using to integers for channel / readout group.
             */
            Pedestal(int channelID, int readoutGroup, float pedestalValue, float pedestalWidth);
            
            /** Convenient constructor using an integer pair for channel / readout group.
             */
            Pedestal(std::pair<int,int> hardwareID, float pedestalValue, float pedestalWidth);
            
            /** Constructor wich only sets channel by two integers. Value and width are set to zero.
             */
            Pedestal(int channelID, int readoutGroup);
            
            /** Constructor which only sets channel by pair of integers. Value and width are set to zero.
             */
            Pedestal(std::pair<int,int> hardwareID);

            /** 'Copy constructor' needed to interpret LCCollection read from file/database.
             */
            Pedestal(EVENT::LCObject* obj):UTIL::LCFixedObject<PedestalNINTVals,
                                                               PedestalNFLOATVals,
                                                               PedestalNDOUBLEVals>(obj) { }

            /** Important for memory handling*/
            virtual ~Pedestal();

                                                               
            // the class interface:
            /** Returns a pair of integers of which the first element ist the channel ID and 
             *  the second element the readout module ID as given by the DAQ.
             */
            std::pair<int,int>  getHardwareID() const;
            
            float getPedestalValue() const;
            float getPedestalWidth() const;

            void setPedestalValue(float pedestalValue);
            void setPedestalWidth(float pedestalWidth);

            static std::string getRevision();
            static std::string getDefaultColName();

            void print(  std::ostream& os = std::cout ) const;

            // -------- need to implement abstract methods from LCGenericObject
            const std::string getTypeName() const
            {
                return std::string("Pedestal");
            }
            const std::string getDataDescription() const
            {
                return std::string("i:ChannelID,ModuleID,f:PedestalValue,PedestalWidth");
            }

            #ifdef USE_LCCD
            // add some helper functions for using the native LCCD conditions map
			typedef std::pair<int, int> key_type;

			key_type conditions_key() const;
			#endif

            
        protected:

            void setHardwareID(int channelID, int readoutGroup);

    };// end of class

    std::ostream &operator<<(std::ostream &os, const Pedestal &p);    

} //end namespace tpcconddata
#endif
